/*
 * analyse_jacknife.c
 * this program analyses the rescue2 results generated by the
 * jacknife_rescue2.sh script
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>

#include "alloc_utils.h"

#define VERSION "1.0"
#define MAX_BUF 10000

char *AA1 = "ARDNCEQGHILKMFPSTWYV";

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  int nclass;              /**< number of classes */
  int shift;               /**< shift number for classes */
  int print_latex_output;  /**< print a LateX output */
  int print_svm_output;    /**< print a SVMTorch-like output */
  int rank;                /**< rank to use for calculating sens/spec */
  int colour;              /**< 0 = red, 1 = blue */
} Parameters;

typedef struct data
{
  int **res;  /**< results counts: in res[i][j], i is the true class
                   and j the predicted class. */
  int *TP;    /**< number of true positives */
  int *FP;    /**< number of false positives */
  int *FN;    /**< number of false negatives */
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->nclass = 20;
  params->shift = 0;
  params->print_latex_output = 0;
  params->print_svm_output = 0;
  params->rank = 1;
  params->colour = 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv)
{
  fprintf(fp, "\nUsage: %s [options] < results_file\n\n", basename(argv[0]));
  fprintf(fp, "  Options:\n\n");
  fprintf(fp, "    -c {0,1}      : 0 = red, 1 = blue for the LateX output\n\n");
  fprintf(fp, "    -h, --help    : print this help\n\n");
  fprintf(fp, "    -nc int       : number of classes.\n\n");
  fprintf(fp, "    -pl           : print a LateX output.\n\n");
  fprintf(fp, "    -psvm         : print a SVMTest-like output.\n\n");
  fprintf(fp, "    -r int        : max rank to use to compute sens/spec "
      "(default = 1).\n\n");
  fprintf(fp, "    -sh int       : shift between the first class and 0.\n\n");
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void read_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-c") == 0) {

/* -c : colour */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-c' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      params->colour = atoi(argv[i]);
      if (params->colour != 0 && params->colour != 1) {
        fprintf(stderr, "Error: colour should be 0 or 1.\n");
        print_usage(stderr, argv);
        exit(1);
      }
    } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help */

      print_usage(stdout, argv);
      exit(0);
    } else if (strcmp(argv[i], "-nc") == 0) {

/* -nc : number of classes */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-nc' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      params->nclass = atoi(argv[i]);
    } else if (strcmp(argv[i], "-pl") == 0) {

/* -pl : print LateX output */

      params->print_latex_output = 1;
    } else if (strcmp(argv[i], "-psvm") == 0) {

/* -psvm : print a SVMTorch like output */

      params->print_svm_output = 1;
    } else if (strcmp(argv[i], "-r") == 0) {

/* -r : maximum rank to use for finding the good prediction */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-r' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      params->rank = atoi(argv[i]);
      if (params->rank <= 0) {
        fprintf(stderr, "Error: rank value should be positive.\n");
        print_usage(stderr, argv);
        exit(1);
      }
    } else if (strcmp(argv[i], "-sh") == 0) {

/* -sh : shift to 0 */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-sh' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      params->shift = atoi(argv[i]);
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else {

/* Unknown tag */

      fprintf(stderr, "Error: unknown tag: '%s'\n", argv[i]);
      print_usage(stderr, argv);
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * NOTE: the only thing the RANK option does is to treat a good result found
 *       at the Nth rank as if it was at the first rank.
 */
void read_results(Data *data, Parameters *params)
{
  char buffer[MAX_BUF], *sp;
  int i, j, k, l, line;

  CALLOC2(data->res, params->nclass, params->nclass, int);
  CALLOC(data->TP, params->nclass + 1, int);
  CALLOC(data->FP, params->nclass + 1, int);
  CALLOC(data->FN, params->nclass + 1, int);

  line = 0;
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    sscanf(buffer + 10, "%d", &i);
    fgets(buffer, MAX_BUF, stdin);
    sscanf(buffer + 8, "%d", &j);

/* begin of RANK option *******************************************************/
    if (params->rank > 1) {
      sp = strtok(buffer + 8, " ");
      for (l = 0; l < params->rank; l++) {
        if (sp == NULL) {
          fprintf(stderr, "Error: not enought data for rank %d at line %d.\n",
              params->rank, line + 2);
          exit(1);
        }
        k = atoi(sp);
        if (k == i) {
          j = i;
          break;
        }
        sp = strtok(NULL, " \n");
      }
    }
/* end of RANK option *********************************************************/

    line += 2;
    if ((i - params->shift) >= params->nclass ||
        (j - params->shift) >= params->nclass) {
      fprintf(stderr, "Error: class out of bound (i - shift = %d, "
          "j - shift = %d) with a maximum of %d.\n", i - params->shift,
          j - params->shift, params->nclass - 1);
      fprintf(stderr, "       line %d:\n%s", line, buffer);
      exit(1);
    }
    data->res[i - params->shift][j - params->shift]++;
  }

  for (i = 0; i < params->nclass; i++) {
    for (j = 0; j < params->nclass; j++) {
      if (i == j) {
        data->TP[i] += data->res[i][i];
      } else {
        data->FP[j] += data->res[i][j];
        data->FN[i] += data->res[i][j];
      }
    }
  }

  for (i = 0; i < params->nclass; i++) {
    data->TP[params->nclass] += data->TP[i];
    data->FP[params->nclass] += data->FP[i];
    data->FN[params->nclass] += data->FN[i];
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print results in text format
 */
void print_sens_spec(Data *data, Parameters *params)
{
  int i, div1, div2, div3;

  for (i = 0; i < params->nclass; i ++) {
    div1 = data->TP[params->nclass] + data->FN[params->nclass];
    if (div1 == 0) {
      div1 = 1;
    }
    div2 = data->TP[i] + data->FN[i];
    if (div2 == 0) {
      div2 = 1;
    }
    div3 = data->TP[i] + data->FP[i];
    if (div3 == 0) {
      div3 = 1;
    }
    printf("Classe %3d (%6.2f) : true = %5d, obs = %5d, TP = %5d : "
        "sens = %8.2f %%, FP = %5d : spec = %8.2f %%\n", i + params->shift,
        100.0 * (float) (data->TP[i] + data->FN[i]) / div1 ,
        data->TP[i] + data->FN[i], data->TP[i] + data->FP[i],
        data->TP[i], 100.0 * (float) data->TP[i] / div2,
        data->FP[i], 100.0 * (float) data->TP[i] / div3);
  }
  div1 = data->TP[params->nclass] + data->FN[params->nclass];
  if (div1 == 0) {
    div1 = 1;
  }
  div2 = data->TP[params->nclass] + data->FP[params->nclass];
  if (div2 == 0) {
    div2 = 1;
  }
  printf("Total      (100.00) : true = %5d, obs = %5d, TP = %5d : "
      "sens = %8.2f %%, FP = %5d : spec = %8.2f %%\n",
      data->TP[params->nclass] + data->FN[params->nclass],
      data->TP[params->nclass] + data->FP[params->nclass],
      data->TP[params->nclass],
      100.0 * (float) data->TP[params->nclass] / div1,
      data->FP[params->nclass],
      100.0 * (float) data->TP[params->nclass] / div2);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print results in SVMTorch-like format
 */
void print_svm_res(Data *data, Parameters *params)
{
  int i, j, div1, div2;
  FILE *fp;

  FOPEN(fp, "svm.out", "w");

  for (i = 0; i < params->nclass; i++) {
    fprintf(fp, "# Classe %d\n", i);
    fprintf(fp, "# Number of points in class : %7d\n",
        data->TP[i] + data->FP[i]);
    fprintf(fp, "# Number of expected points : %7d\n",
        data->TP[i] + data->FN[i]);
    fprintf(fp, "  -> True positives         : %7d\n", data->TP[i]);
    fprintf(fp, "  -> False positives        : %7d", data->FP[i]);
    for (j = 0; j < params->nclass; j++) {
      fprintf(fp, " %d:[%5d]", j, data->res[j][i]);
    }
    div1 = data->TP[i] + data->FN[i];
    if (div1 == 0) {
      div1 = 1;
    }
    div2 = data->TP[i] + data->FP[i];
    if (div2 == 0) {
      div2 = 1;
    }
    fprintf(fp, "\n  -> Sensitivity            : %6.2f%%\n",
        100.0 * (float) data->TP[i] / div1);
    fprintf(fp, "  -> Specificity            : %6.2f%%\n\n",
        100.0 * (float) data->TP[i] / div2);
  }

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define YSIZE 4.0
#define XSTEP 0.1
#undef PLOT_OBSERVED_FREQUENCIES

/**
 * Print results in LateX format
 */
void print_latex_output(Data *data, Parameters *params)
{
  int i, div1, div2;
  FILE *fp;

  FOPEN(fp, "analyse_jacknife.tex", "w");

  fprintf(fp, "\\begin{pspicture}(%.2f,%.2f)(%.2f,%.2f)\n", 0.0, -1.0,
      6.0 * 20 * XSTEP, YSIZE);
  fprintf(fp, "\\psline(0,0)(0,%.2f)\n", YSIZE);
  fprintf(fp, "\\psline(%.2f,0)(%.2f,%.2f)\n", 6.0 * XSTEP * 20,
      6.0 * XSTEP * 20, YSIZE);
  fprintf(fp, "\\psline(0,%.2f)(%.2f,%.2f)\n", YSIZE / 4.0, 6.0 * XSTEP * 20,
      YSIZE / 4.0);
  fprintf(fp, "\\psline(0,%.2f)(%.2f,%.2f)\n", YSIZE / 2.0,
      6.0 * XSTEP * 20, YSIZE / 2.0);
  fprintf(fp, "\\psline(0,%.2f)(%.2f,%.2f)\n", 3.0 * YSIZE / 4.0,
      6.0 * XSTEP * 20, 3.0 * YSIZE / 4.0);
  fprintf(fp, "\\psline(0,%.2f)(%.2f,%.2f)\n", YSIZE, 6.0 * XSTEP * 20, YSIZE);
  fprintf(fp, "\\rput[r](-0.5,%.2f){25\\%%}\n", YSIZE / 4.0);
  fprintf(fp, "\\rput[r](-0.5,%.2f){50\\%%}\n", YSIZE / 2.0);
  fprintf(fp, "\\rput[r](-0.5,%.2f){75\\%%}\n", 3.0 * YSIZE / 4.0);
  fprintf(fp, "\\rput[r](-0.5,%.2f){100\\%%}\n", YSIZE);
  for (i = 0; i < params->nclass; i ++) {
    div1 = data->TP[i] + data->FN[i];
    if (div1 == 0) {
      div1 = 1;
    }
    div2 = data->TP[i] + data->FP[i];
    if (div2 == 0) {
      div2 = 1;
    }
    if (params->colour == 0) {
      fprintf(fp, "\\rput(%.2f,%.2f){%c}\n", 6.0 * i * XSTEP + 3.0 * XSTEP,
          -0.5, AA1[i]);
#ifdef PLOT_OBSERVED_FREQUENCIES
      fprintf(fp, "\\psframe*(%.2f,%.2f)(%.2f,%.2f)\n", 6.0 * i * XSTEP, 0.0,
          6.0 * i * XSTEP + 2.0 * XSTEP,
          YSIZE * (float) data->TP[i] + data->FN[i] /
          (data->TP[params->nclass] + data->FN[params->nclass]));
#endif
      fprintf(fp, "\\psframe[fillstyle=solid,fillcolor=lightlightred,"
          "linecolor=lightlightred](%.2f,%.2f)(%.2f,%.2f)\n",
          6.0 * i * XSTEP + 2.0 * XSTEP, 0.0, 6.0 * i * XSTEP + 3.0 * XSTEP,
          YSIZE * (float) data->TP[i] / div1);
      fprintf(fp, "\\psframe[fillstyle=solid,fillcolor=red,linecolor=red]"
          "(%.2f,%.2f)(%.2f,%.2f)\n", 6.0 * i * XSTEP + 3.0 * XSTEP, 0.0,
          6.0 * i * XSTEP + 4.0 * XSTEP, YSIZE * (float) data->TP[i] / div2);
    } else {
      fprintf(fp, "\\psframe[fillstyle=solid,fillcolor=lightlightblue,"
          "linecolor=lightlightblue](%.2f,%.2f)(%.2f,%.2f)\n",
          6.0 * i * XSTEP + 4.0 * XSTEP, 0.0, 6.0 * i * XSTEP + 5.0 * XSTEP,
          YSIZE * (float) data->TP[i] / div1);
      fprintf(fp, "\\psframe[fillstyle=solid,fillcolor=blue,linecolor=blue]"
          "(%.2f,%.2f)(%.2f,%.2f)\n", 6.0 * i * XSTEP + 5.0 * XSTEP, 0.0,
          6.0 * i * XSTEP + 6.0 * XSTEP, YSIZE * (float) data->TP[i] / div2);
    }
  }
  fprintf(fp, "\\end{pspicture}\n");

  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  Parameters params;
  Data data;

  init_parameters(&params);
  read_command_line(argc, argv, &params);

  read_results(&data, &params);

  print_sens_spec(&data, &params);
  if (params.print_svm_output == 1) {
    print_svm_res(&data, &params);
  }
  if (params.print_latex_output == 1) {
    print_latex_output(&data, &params);
  }

  return 0;
}
