#!/bin/bash

# jacknife_rescue2.sh
# This script is used to jackknife the rescue2 program on a given set of
# BMRB files.
#
# Copyright (C) 2003 Antoine MARIN
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License, version 2.1 published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# You can contact the author via mail or email at:
# Antoine Marin (Antoine.Marin@ibpc.fr)
# CNRS UPR 9080
# Laboratoire de Biochimie Thorique
# 13, rue Pierre et Marie Curie
# 75005 Paris - France

################################################################################
#
################################################################################
#
# Default values
#
VERSION="1.0"
# Directories
BMRB_DIR="${HOME}/BMRB"
# Files
SEQ_LIST="BMRB_selected_sequences.txt"
INPUT_CS="cs.in"
INPUT_STATS="BMRB_stats.out"
# Programs
BMRB_STATS="${HOME}/cvs_work/BMRB_tools/BMRB_stats"
BMRB_TO_CS="${HOME}/cvs_work/BMRB_tools/BMRB_to_cs"
RESCUE2="${HOME}/cvs_work/RESCUE2/rescue2"

################################################################################
#
################################################################################
print_help()
{
  echo ""
  echo "Usage: ${0##*/} [options]"
  echo ""
  echo "  Options:"
  echo ""
  echo "    --clean       : remove generated files"
  echo ""
  echo "    -d directory  : directory containing BMRB files"
  echo "                    (default = $BMRB_DIR)"
  echo ""
  echo "    -h, --help    : print this help"
  echo ""
  echo "    -i file       : input file containing the BMRB files list"
  echo "                    (default = $SEQ_LIST)"
  echo ""
  echo "    -v, --version : print version only (current version = $VERSION)"
  echo ""
}

################################################################################
#
################################################################################
clean()
{
  \rm -f $LEARN_LIST $STATS_FILE $CS_FILE $RES_FILE1 $RES_FILE2
}

################################################################################
#
################################################################################
#
# Process command line
#
while test -n "`echo $1 | grep '-'`"; do
  case $1 in
    "--clean") clean; exit 0 ;;
    "-d") BMRB_DIR=$2; shift ;;
    "-h"|"--help") print_help; exit 0 ;;
    "-i") SEQ_LIST=$2; shift ;;
    "-v"|"--version") echo "This is `basename $0` release number $VERSION"; exit 0 ;;
    *) echo "Error: unknowk tag '$1'"; print_help; exit 1 ;;
  esac
  shift
done

################################################################################
#
################################################################################
#
# Check some variables
#
flag=0
if test ! -d $BMRB_DIR; then
  echo "**** Directory '$BMRB_DIR' does not exist"
  flag=1
fi

if test ! -f $SEQ_LIST; then
  echo "**** File '$SEQ_LIST' does not exist"
  flag=1
fi

if test ! -f $INPUT_CS; then
  echo "**** File '$INPUT_CS' does not exist"
  flag=1
fi

if test ! -f $INPUT_STATS; then
  echo "**** File '$INPUT_STATS' does not exist"
  flag=1
fi

if test $flag = "1"; then
  echo "**** Errors occured ****"
  print_help
  exit 1
fi

################################################################################
#
################################################################################
#
# Set file names
#
LEARN_LIST="${SEQ_LIST%.*}.learn"
STATS_FILE="${SEQ_LIST%.*}.stats"
CS_FILE="${SEQ_LIST%.*}.xml"
RES_FILE1="${SEQ_LIST%.*}.results.indep"
RES_FILE2="${SEQ_LIST%.*}.results.dep"

################################################################################
#
################################################################################
#
# Jackknife procedure
#
num_seq=`\cat $SEQ_LIST | wc -l`

if test -f $RES_FILE1 || test -f $RES_FILE2; then
  echo "Resutls file '$RES_FILE1' or '$RES_FILE2' is on the way ... remove it first"
  exit 0
fi

i=1
for file in `cat $SEQ_LIST`; do

# Extract chemical shifts
  $BMRB_TO_CS -f -ics $INPUT_CS -ist $INPUT_STATS -mc 0 < ${BMRB_DIR}/$file > $CS_FILE

# test if file is empty
  if test -s $CS_FILE; then
    echo "Testing sequence '$file' ($i)"

# Learning file
    grep -v $file $SEQ_LIST > $LEARN_LIST

# Compute statistics
    $BMRB_STATS -bd $BMRB_DIR -i $INPUT_CS -o $STATS_FILE < $LEARN_LIST

# Calculate proba
    $RESCUE2 -af $INPUT_CS -sf $STATS_FILE -cspi < $CS_FILE >> $RES_FILE1
    $RESCUE2 -af $INPUT_CS -sf $STATS_FILE -cspd < $CS_FILE >> $RES_FILE2
  fi

  i=`echo "$i + 1" | bc`
done

# Clean
\rm -f $LEARN_LIST $STATS_FILE $CS_FILE
