/*
 * rescue2.c
 * This program calculates the probabilities that a given spin system
 * comes from any of the 20 amino acid types.
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <libgen.h> /* for basename */
#include <search.h> /* for hash tables */
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_combination.h>
#include <gsl/gsl_blas.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_vector.h>

#include "alloc_utils.h"

#define VERSION "1.0.2"

#define MAX_BUF 1024
#define MAX_RES 20
#define MAX_CS 50
#define EPSILON 1e-5
#define MAX_NAME_LEN 16
#define M_PI 3.14159265358979323846 /* pi */
#define PROBA_CUT_OFF 1e-6

char *AA1 = "ARDNCEQGHILKMFPSTWYV";

char *AA3[] = {"ALA", "ARG", "ASP", "ASN", "CYS", "GLU", "GLN", "GLY", "HIS",
 "ILE", "LEU", "LYS", "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "VAL"
};

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  char *BMRB_stats_fname;   /**< input file containing statistics on the
                                 Bio-Mag-Res-Bank (BMRB) */
  char *input_atoms_fname;  /**< input file containing atoms to select */
  int full_perm_indep;      /**< if = 1, calculate full permutations
                                 independant probabilities */
  int full_perm_dep;        /**< if = 1, calculate full permutations
                                 dependant probabilities (using multivariate
                                 normal distributions) */
  int select_perm_indep;    /**< if = 1, calculate selected permutations
                                 (the most probable) independant
                                 probabilities */
  int select_perm_dep;      /**< if = 1, calculate selected permutations
                                 (the most probable) dependant probabilities
                                 (using multivariate normal distributions) */
  int verbosity;
} Parameters;

typedef struct enum_data {
  int **val;     /**< each column contains the permited values */
  int *max_val;  /**< maximum number of values for each column in 'val' */
  int *pos;      /**< index of current used positions */
  int *tag;      /**< tags for used distributions */
} Enum_data;

/**
 * Stat_data contains statistics about a spin sytem (a class)
 */
typedef struct stat_data {
  int nH;           /**< number of H_val */
  int nC;           /**< number of C_val */
  int nN;           /**< number of N_val */
  int n_total;      /**< nH + nC + nN */
  double *mean;     /**< chemical shifts means */
  double *std_dev;  /**< chemical shifts standard deviations */
  double *miss;     /**< percentage of missing chemical shifts */
  double **cov;     /**< covariances matrix */
  char *atom_type;  /**< atom types for chemical shift */
} Stat_data;

/**
 * Val_data contains input values
 */
typedef struct val_data {
  int class;      /**< spin system class */
  int nH;         /**< number of H_val */
  int nC;         /**< number of C_val */
  int nN;         /**< number of N_val */
  int n_total;    /**< nH + nC + nN */
  double *H_val;  /**< Hydrogen chemical shifts values */
  double *C_val;  /**< carbon chemical shifts values */
  double *N_val;  /**< Azote chemical shifts values */
  double *proba;  /**< probabilities for each class */
} Val_data;

typedef struct data {
  char **cs_keys;    /**< keys of the input types for the hash table */
  int *res_num;      /**< residue number for each input type */
  int *cs_num;       /**< chemical shift number for each input type
                          of each residue */
  int num_res;       /**< number of residues = number of classes */
  int num_keys;      /**< number of different keys */
  int num_val;       /**< number of values in vdata */
  Stat_data *sdata;  /**< statistic parameters for each class */
  Val_data *vdata;   /**< input chemical shifts values */
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->BMRB_stats_fname = NULL;
  params->input_atoms_fname = NULL;
  params->full_perm_indep = 0;
  params->full_perm_dep = 0;
  params->select_perm_indep = 0;
  params->select_perm_dep = 0;
  params->verbosity = 2;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv)
{
  fprintf(fp, "\nUsage: %s [options] -sf BMRB_stats_file -af input_atoms_file "
      "< cs_xml_file\n\n", basename(argv[0]));
  fprintf(fp, "  Options:\n\n");
  fprintf(fp, "    -cfpd         : calculate full permutation dependant "
      "probabilities.\n\n");
  fprintf(fp, "    -cfpi         : calculate full permutation independant "
      "probabilities.\n\n");
  fprintf(fp, "    -cspd         : calculate selected permutation dependant "
      "probabilities.\n\n");
  fprintf(fp, "    -cspi         : calculate selected permutation "
      "independant probabilities.\n\n");
  fprintf(fp, "    -h, --help    : print this help\n\n");
  fprintf(fp, "    -af file      : file containing atoms to analyse "
      "(mandatory).\n");
  fprintf(fp, "                    file format is:\n");
  fprintf(fp, "                    # Amino   Atom    Atom\n");
  fprintf(fp, "                    # Acid    Name    Type\n");
  fprintf(fp, "                    ALA     HA       H\n");
  fprintf(fp, "                    ALA     CB       C\n");
  fprintf(fp, "                    ...     ...      ...\n\n");
  fprintf(fp, "   -pb            : don't print the best class.\n");
  fprintf(fp, "   +pb            : print the best class (this is the default).\n\n");
  fprintf(fp, "   -sf file       : enhanced BMRB statistics file generated by the\n"
      "                    'BMRB_stats' program from A.Marin (mandatory).\n\n");
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
  fprintf(fp, "NOTE: cs_xml_file is the file containing the chemical shifts\n"
      "      grouped by spin system in XML, generated by the 'BMRB_to_cs'\n"
      "      program from A.Marin (mandatory).\n\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void process_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {
    if (strcmp(argv[i], "-cfpd") == 0) {

/* -cfpd : calculate full permutation dependant probabilities */

      params->full_perm_dep = 1;
    } else if (strcmp(argv[i], "-cfpi") == 0) {

/* -cfpi : calculate full permutation independant probabilities */

      params->full_perm_indep = 1;
    } else if (strcmp(argv[i], "-cspd") == 0) {

/* -cspd : calculate selected permutation dependant probabilities */

      params->select_perm_dep = 1;
    } else if (strcmp(argv[i], "-cspi") == 0) {

/* -cspi : calculate selected permutation independant probabilities */

      params->select_perm_indep = 1;
    } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help */

      print_usage(stdout, argv);
      exit(0);
    } else if (strcmp(argv[i], "-af") == 0) {

/* -af : atoms file */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-af' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      ALLOC(params->input_atoms_fname, strlen(argv[i]) + 1, char);
      strcpy(params->input_atoms_fname, argv[i]);
    } else if (strcmp(argv[i], "-sf") == 0) {

/* -sf : statistics file */

      i++;
      if (i >= argc) {
        fprintf(stderr, "Error: you should provide a value for the "
            "'-sf' option.\n");
        print_usage(stderr, argv);
        exit(1);
      }
      ALLOC(params->BMRB_stats_fname, strlen(argv[i]) + 1, char);
      strcpy(params->BMRB_stats_fname, argv[i]);
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else {

/* Unknown tag */

      fprintf(stderr, "Error: unknown tag: '%s'\n", argv[i]);
      print_usage(stderr, argv);
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Setup hash table for amino acid types
 */
void init_htbl_AA3()
{
  int i;
  ENTRY ent, *ep;

  for (i = 0; i < 20; i++) {
    ent.key = AA3[i];

/*
 * -2 is important because this value is compared to the last residue in
 * function 'read_stats' and at the begining, the last residue is
 * initialized to -1 and it has to be different !
 */

    ent.data = (char *) -2;
    ep = hsearch(ent, ENTER);
    if (ep == NULL) {
      fprintf(stderr, "Error while building hash table for AA3.\n");
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Read chemical shifts to analyze
 */
void read_cs_to_analyse(Parameters *params, Data *data)
{
  char buffer[MAX_BUF], amino_acid[MAX_NAME_LEN];
  char atom_name[MAX_NAME_LEN], atom_type[MAX_NAME_LEN];
  FILE *fp;
  int i, ires, itype, lines, last_res;
  ENTRY ent, *ep;
  Stat_data *sdata;

/* get number of lines */

  FOPEN(fp, params->input_atoms_fname, "r");

  lines = 0;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (buffer[0] != '#' && buffer[0] != '\n' &&
        sscanf(buffer, "%s%s%s", amino_acid, atom_name, atom_type) == 3) {
      lines++;
    }
  }
  if (lines == 0) {
    fprintf(stderr, "Error: there is no atom types to analyse in "
        "file '%s'.\n", params->input_atoms_fname);
    exit(1);
  }

/* allocate memory to store hash keys */

  ALLOC2(data->cs_keys, lines, 3 * MAX_NAME_LEN, char);
  ALLOC(data->res_num, lines, int);
  ALLOC(data->cs_num, lines, int);

/* init hash table */

  hcreate(4 * lines + MAX_RES);
  init_htbl_AA3();

/* get keys and fill the hash table */

  rewind(fp);
  i = itype = 0;
  last_res = ires = -1;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (buffer[0] != '#' && buffer[0] != '\n' &&
        sscanf(buffer, "%s%s%s", amino_acid, atom_name, atom_type) == 3) {

/*  check if the type is not alrady in the hash table */

      sprintf(data->cs_keys[i], "%s %s %s", amino_acid, atom_name,
          atom_type);
      ent.key = data->cs_keys[i];
      ep = hsearch(ent, FIND);
      if (ep == NULL) {

/* change data associated with amino acid type in hash table */

        ent.key = amino_acid;
        ep = hsearch(ent, FIND);
        if (ep != NULL) {
          if (last_res != (int) ep->data) {
            itype = 0;
            ires++;
            ep->data = (char *) ires;
            last_res = ires;
          } else {
            itype++;
          }
          data->res_num[i] = ires;
          data->cs_num[i] = itype;
        } else {
          fprintf(stderr, "Error: unknown residue '%s' in file '%s'.\n",
              amino_acid, params->input_atoms_fname);
          exit(1);
        }

/* add type in hash table */

        ent.key = data->cs_keys[i];
        ent.data = (char *) itype;
        ep = hsearch(ent, ENTER);
        if (ep == NULL) {
          fprintf(stderr, "Error while building hash table for input types.\n");
          exit(1);
        }
        i++;
      }
    }
  }

/* if there is no redundancy 'i' should be equal to 'lines' */

  if (i != lines) {
    fprintf(stderr, "Warning: the number of different entries (%d) is not "
        "the same as the number of input types (%d) in file '%s'\n",
        i, lines, params->input_atoms_fname);
  }
  data->num_keys = i;
  data->num_res = ires + 1;
  CALLOC(data->sdata, data->num_res, Stat_data);

/* calculate number of residues = number of classes */

  for (i = 0; i < data->num_keys; i++) {
    sdata = &data->sdata[data->res_num[i]];
    sdata->n_total++;
  }

  FCLOSE(fp);

#ifdef DEBUG
  printf("Number of keys = %d\n", data->num_keys);
  printf("Keys:\n");
  for (i = 0; i < data->num_keys; i++) {
    printf("%s\n", data->cs_keys[i]);
  }
  printf("\n");
  printf("Residue index (for each key):\n");
  for (i = 0; i < data->num_keys; i++) {
    printf("%d\n", data->res_num[i]);
  }
  printf("\n");
  printf("Chemical shift index (for each key):\n");
  for (i = 0; i < data->num_keys; i++) {
    printf("%d\n", data->cs_num[i]);
  }
  printf("\n");
  printf("Number of residues = %d\n", data->num_res);
  printf("Number of chemical shifts for each residue:\n");
  for (i = 0; i < data->num_res; i++) {
    sdata = &data->sdata[i];
    printf("%d\n", sdata->n_total);
  }
  printf("\n");
#endif
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Read BMRB statistics
 */
void read_stats(Parameters *params, Data *data)
{
  char buffer[MAX_BUF], amino_acid[MAX_NAME_LEN], atom_desc[MAX_NAME_LEN];
  char atom_name[MAX_NAME_LEN], atom_type[MAX_NAME_LEN], *sp;
  FILE *fp;
  int i, j, k, jj, kk, ires, last_ires, itype, num_shifts;
  int num_missings, **tags;
  double min_shift, max_shift, mean_shift, stdev_shift, percentage_missing;
  ENTRY ent, *ep;
  Stat_data *sdata;

/* read corresping values into the enhanced BMRB stats file */

  FOPEN(fp, params->BMRB_stats_fname, "r");

  if (fgets(buffer, MAX_BUF, fp) == NULL) {
    fprintf(stderr, "Error: unexpected End-Of-File in '%s'\n",
        params->BMRB_stats_fname);
    exit(1);
  }
  if (fgets(buffer, MAX_BUF, fp) == NULL) {
    fprintf(stderr, "Error: unexpected End-Of-File in '%s'\n",
        params->BMRB_stats_fname);
    exit(1);
  }

  CALLOC2(tags, MAX_RES, MAX_CS, int);

/* allocate memory for classes */

  for (i = 0; i < data->num_res; i++) {
    sdata = &data->sdata[i];
    ALLOC(sdata->mean, sdata->n_total, double);
    ALLOC(sdata->std_dev, sdata->n_total, double);
    ALLOC(sdata->miss, sdata->n_total, double);
    ALLOC2(sdata->cov, sdata->n_total, sdata->n_total, double);
    ALLOC(sdata->atom_type, sdata->n_total, char);
  }

/* get means, std devs and missings values */

  i = 2;
  last_ires = -1;
  k = 0;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (strncmp(buffer, "<cov ", 5) == 0) {
      break;
    }
    i++;
    if (sscanf(buffer, "%s%s%s%d%lf%lf%lf%lf%d%lf", amino_acid, atom_name,
        atom_type, &num_shifts, &min_shift, &max_shift, &mean_shift,
        &stdev_shift, &num_missings, &percentage_missing) == 10) {

/* find AA */

      ent.key = amino_acid;
      ep = hsearch(ent, FIND);
      if (ep == NULL) {
        fprintf(stderr, "Error: unknown amino acid (%s) in stat file "
            "'%s' line %d\n", amino_acid, params->BMRB_stats_fname, i);
        fprintf(stderr, "%s", buffer);
        exit(1);
      }
      if ((int) ep->data < 0) {
        fprintf(stderr, "Error: amino acid '%s' in stat file '%s' "
            "has not been selected in '%s' file.\n", amino_acid,
            params->BMRB_stats_fname, params->input_atoms_fname);
        exit(1);
      }
      ires = (int) ep->data;
      if (ires != last_ires) {
        last_ires = ires;
        k = 0;
      }

/* find description */

      sprintf(atom_desc, "%s %s %s", amino_acid, atom_name, atom_type);
      ent.key = atom_desc;
      ep = hsearch(ent, FIND);
      if (ep != NULL) {
        itype = (int) ep->data;
        tags[ires][k] = 1;
        sdata = &data->sdata[ires];
        sdata->mean[itype] = mean_shift;
        sdata->std_dev[itype] = stdev_shift;
        sdata->miss[itype] = percentage_missing / 100.0;
        sdata->atom_type[itype] = atom_type[0];
        if (atom_type[0] == 'H') {
          sdata->nH++;
        } else if (atom_type[0] == 'C') {
          sdata->nC++;
        } else if (atom_type[0] == 'N') {
          sdata->nN++;
        } else {
          fprintf(stderr, "Error: unknown atom type '%c' "
              "(should be H, C or N) in file '%s'.\n", atom_type[0],
              params->BMRB_stats_fname);
          exit(1);
        }
      }
      k++;
    }
  }

/* get covariances */

  rewind(fp);
  i = 0;
  j = 0;
  while (fgets(buffer, MAX_BUF, fp) != NULL) {
    if (strncmp(buffer, "<cov ", 5) == 0) {

/* get residue number to point on the right Stat_data */

      sp = strstr(buffer, "res=");
      sp += 5;
      strncpy(amino_acid, sp, 3);
      amino_acid[3] = '\0';
      ent.key = amino_acid;
      ep = hsearch(ent, FIND);
      if (ep == NULL) {
        fprintf(stderr, "Error: unknown amino acid (%s) in stat file "
            "'%s' while reading covariances\n", amino_acid,
            params->BMRB_stats_fname);
        fprintf(stderr, "%s", buffer);
        exit(1);
      }
      if ((int) ep->data < 0) {
        fprintf(stderr, "Error: amino acid '%s' in stat file '%s' "
            "has not been selected in '%s' file.\n", amino_acid,
            params->BMRB_stats_fname, params->input_atoms_fname);
        exit(1);
      }
      ires = (int) ep->data;
      sdata = &data->sdata[ires];

/* get dimension to scan all values */

      sp = strstr(buffer, "dim=");
      sp += 5;
      itype = atoi(sp);
      if (fgets(buffer, MAX_BUF, fp) == NULL) {
        fprintf(stderr, "Error: unexpected End-Of-File in '%s'\n",
            params->BMRB_stats_fname);
        exit(1);
      }

      j = 0;
      jj = 0;
      while (strncmp(buffer, "</cov>", 6) != 0) {

/*
 * only get values that were read before (tags[][] == 1)
 * ATTENTION: i assume the same order for chemical shifts
 */

        if (tags[ires][j] == 1) {
          sp = strtok(buffer, " \n");
          kk = 0;
          for (k = 0; k < itype; k++) {
            if (tags[ires][k] == 1) {
              sdata->cov[jj][kk] = atof(sp);
              kk++;
            }
            sp = strtok(NULL, " \n");
          }
          jj++;
        }
        if (fgets(buffer, MAX_BUF, fp) == NULL) {
          fprintf(stderr, "Error: unexpected End-Of-File in '%s'\n",
              params->BMRB_stats_fname);
          exit(1);
        }
        j++;
      }
      i++;
    }
  }

  FREE2(tags);
  FCLOSE(fp);

#ifdef DEBUG
  for (i = 0; i < data->num_res; i++) {
    printf("Class %d\n", i);
    sdata = &data->sdata[i];
    printf("  nH =%3d, nC =%3d, nN =%3d, total =%3d\n", sdata->nH,
        sdata->nC, sdata->nN, sdata->n_total);
    printf("  means:\n");
    for (itype = 0; itype < sdata->n_total; itype++) {
      printf("%8.2f", sdata->mean[itype]);
    }
    printf("\n");
    printf("  std devs:\n");
    for (itype = 0; itype < sdata->n_total; itype++) {
      printf("%8.2f", sdata->std_dev[itype]);
    }
    printf("\n");
    printf("  missings:\n");
    for (itype = 0; itype < sdata->n_total; itype++) {
      printf("%8.2f", sdata->miss[itype]);
    }
    printf("\n");
    printf("  atom types:\n");
    for (itype = 0; itype < sdata->n_total; itype++) {
      printf("%8c", sdata->atom_type[itype]);
    }
    printf("\n");
    printf("  covariances:\n");
    for (j = 0; j < sdata->n_total; j++) {
      for (k = 0; k < sdata->n_total; k++) {
        printf("%8.2f", sdata->cov[j][k]);
      }
      printf("\n");
    }
  }
  printf("\n");
#endif
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Read input values to predict
 */
void read_input_values(Data *data)
{
  int i, num_res, num_cs;
  char buffer[MAX_BUF], cs_type, *sp, amino_acid[MAX_NAME_LEN];
  Val_data *vdata;
  double *val;
  ENTRY ent, *ep;

/* count residues */

  num_res = 0;
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (strncmp(buffer, "<res ", 5) == 0) {
      num_res++;
    }
  }
  data->num_val = num_res;

/* allocate memory */

  CALLOC(data->vdata, num_res, Val_data);

/* get values */

  rewind(stdin);
  num_res = 0;
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (strncmp(buffer, "<res ", 5) == 0) {
      vdata = &data->vdata[num_res];
      CALLOC(vdata->proba, data->num_res, double);

/* get AA */

      sp = strstr(buffer, "type");
      sp += 6;
      strncpy(amino_acid, sp, 3);
      amino_acid[3] = '\0';
      ent.key = amino_acid;
      ep = hsearch(ent, FIND);

/*
 * if ep == NULL, the amino acid is not known and the prediction will
 * be available in the form of probabilities, if ep != NULL, the program
 * is used for an evaluation purpose and the results will be available
 * in the form of sorted classes.
 */

      if (ep == NULL) {
        vdata->class = -1;
      } else {
        if ((int) ep->data < 0) {
          fprintf(stderr, "Warning: no statistics have been read for this "
              "residue ('%s' number %d in input file).", amino_acid,
              num_res + 1);
          continue;
        }
        vdata->class = (int) ep->data;
      }

      if (fgets(buffer, MAX_BUF, stdin) == NULL) {
        fprintf(stderr, "Error: unexpected End-Of-File while reading stdin\n");
        exit(1);
      }
      while (strncmp(buffer, "</res>", 6) != 0) {
        if (strncmp(buffer, "  <cs ", 6) == 0) {

/* get atom type */

          sp = strstr(buffer, "type");
          sp += 6;
          cs_type = sp[0];

/* get number of chemical shifts */

          sp = strstr(buffer, "n=");
          sp += 3;
          num_cs = atoi(sp);

/* allocate memory and get values on the next line */

          if (cs_type == 'H') {
            vdata->nH = num_cs;
            ALLOC(vdata->H_val, num_cs, double);
            val = vdata->H_val;
          } else if (cs_type == 'C') {
            vdata->nC = num_cs;
            ALLOC(vdata->C_val, num_cs, double);
            val = vdata->C_val;
          } else if (cs_type == 'N') {
            vdata->nN = num_cs;
            ALLOC(vdata->N_val, num_cs, double);
            val = vdata->N_val;
          } else {
            fprintf(stderr, "Error: unknown chemical shift type '%c'\n",
                cs_type);
            exit(1);
          }
          if (fgets(buffer, MAX_BUF, stdin) == NULL) {
            fprintf(stderr, "Error: unexpected End-Of-File while reading stdin\n");
            exit(1);
          }
          sp = strtok(buffer, " \n");
          for (i = 0; i < num_cs; i++) {
            val[i] = atof(sp);
            sp = strtok(NULL, " \n");
          }
        }
        vdata->n_total = vdata->nH + vdata->nC + vdata->nN;
        if (fgets(buffer, MAX_BUF, stdin) == NULL) {
          fprintf(stderr, "Error: unexpected End-Of-File while reading stdin\n");
          exit(1);
        }
      }
      num_res++;
    }
  }

#ifdef DEBUG
  for (i = 0; i < num_res; i++) {
    vdata = &data->vdata[i];
    printf("Input %6d : H", i);
    for (num_cs = 0; num_cs < vdata->nH; num_cs++) {
      printf("%8.2f", vdata->H_val[num_cs]);
    }
    printf(" C");
    for (num_cs = 0; num_cs < vdata->nC; num_cs++) {
      printf("%8.2f", vdata->C_val[num_cs]);
    }
    printf(" N");
    for (num_cs = 0; num_cs < vdata->nN; num_cs++) {
      printf("%8.2f", vdata->N_val[num_cs]);
    }
    printf(", total = %d\n", vdata->n_total);
  }
  printf("\n");
#endif
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Lower Upper (LU) matrix decomposition
 */
void lu_matrix_decomposition(double **input_mat, int mat_dim, int *index,
    double *det_sign, double **lu_mat)
{
  int i, j, k, piv;
  double val_tmp, *mat_ptr_k, *mat_ptr_i;

  memcpy(lu_mat[0], input_mat[0], mat_dim * mat_dim * sizeof(double));

  *det_sign = 1.0;
  for (i = 0; i < mat_dim; i++) {
    index[i] = i;
  }
  for (k = 0; k < mat_dim; k++) {
    val_tmp = fabs(lu_mat[index[k]][k]);
    for (i = k + 1; i < mat_dim; i++) {
      if (fabs(lu_mat[index[i]][k]) > val_tmp) {
        piv = index[k];
        index[k] = index[i];
        index[i] = piv;
        *det_sign = -(*det_sign);
      }
    }
    mat_ptr_k = lu_mat[index[k]];
    if (mat_ptr_k[k] == 0.0) {
      fprintf(stderr, "Error in lu_matrix_decomposition: singular matrix.\n");
      exit(1);
    }
    val_tmp = mat_ptr_k[k];
    for (i = k + 1; i < mat_dim; i++) {
      mat_ptr_i = lu_mat[index[i]];
      mat_ptr_i[k] /= val_tmp;
      for (j = k + 1; j < mat_dim; j++) {
        mat_ptr_i[j] -= mat_ptr_i[k] * mat_ptr_k[j];
      }
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Lower Upper (LU) forward - backward substitution
 */
void lu_forward_backsubstitution(double **lu_mat, int mat_dim, int *index,
    double *col)
{
  int i, j;
  double sum, *mat_ptr;

/* forward substitution */

  for (i = 0; i < mat_dim; i++) {
    sum = col[i];
    mat_ptr = lu_mat[index[i]];
    for (j = 0; j < i; j++) {
      sum -= mat_ptr[j] * col[j];
    }
    col[i] = sum;
  }

/* backward substitution */

  for (i = mat_dim - 1; i >= 0; i--) {
    sum = col[i];
    mat_ptr = lu_mat[index[i]];
    for (j = i + 1; j < mat_dim; j++) {
      sum -= mat_ptr[j] * col[j];
    }
    col[i] = sum / mat_ptr[i];
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate the determinant of a matrix using LU decomposition
 */
double lu_matrix_determinant(double **input_mat, int mat_dim)
{
  int i, *index;
  double det, **lu_mat;

  ALLOC2(lu_mat, mat_dim, mat_dim, double);
  ALLOC(index, mat_dim, int);

  lu_matrix_decomposition(input_mat, mat_dim, index, &det, lu_mat);
  for (i = 0; i < mat_dim; i++) {
    det *= lu_mat[index[i]][i];
  }

  FREE2(lu_mat);
  FREE(index);

  return det;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate the determinant log of a matrix using LU decomposition
 */
double lu_matrix_log_determinant(double **input_mat, int mat_dim)
{
  int i, *index;
  double det, **lu_mat, tmp_det;

  ALLOC2(lu_mat, mat_dim, mat_dim, double);
  ALLOC(index, mat_dim, int);

  lu_matrix_decomposition(input_mat, mat_dim, index, &det, lu_mat);
  tmp_det = 0.0;
  for (i = 0; i < mat_dim; i++) {
    if (lu_mat[index[i]][i] < 0.0) {
      det *= -1;
      tmp_det += log(-lu_mat[index[i]][i]);
    } else {
      tmp_det += log(lu_mat[index[i]][i]);
    }
  }
  det *= tmp_det;

  FREE2(lu_mat);
  FREE(index);

  return det;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Invert a matrix using LU decomposition
 */
void lu_decomposition_mat_inv(double **input_mat, int mat_dim,
    double **inv_mat)
{
  int i, j, *index;
  double *col, **lu_mat, det_sign;

  ALLOC(col, mat_dim, double);
  ALLOC(index, mat_dim, int);
  ALLOC2(lu_mat, mat_dim, mat_dim, double);

  lu_matrix_decomposition(input_mat, mat_dim, index, &det_sign, lu_mat);
  for (j = 0; j < mat_dim; j++) {
    memset(col, 0, mat_dim * sizeof(double));
    col[j] = 1.0;
    lu_forward_backsubstitution(lu_mat, mat_dim, index, col);
    for (i = 0; i < mat_dim; i++) {
      inv_mat[i][index[j]] = col[i];
    }
  }

  FREE(col);
  FREE(index);
  FREE2(lu_mat);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Print matrix in octave format to file pointer 'fp'
 */
void print_octave_matrix(FILE *fp, double **mat, int mat_dim)
{
  int i, j;

  fprintf(fp, "[");
  for (i = 0; i < mat_dim; i++) {
    fprintf(fp, "[");
    for (j = 0; j < mat_dim; j++) {
      fprintf(fp, "%10.4g", mat[i][j]);
    }
    fprintf(fp, "];");
  }
  fprintf(fp, "]\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/*
 * Mahalanobis distance = (X - M)T . A-1 . (X - M)
 * X = observed values
 * M = mean values
 * A-1 = inverted covariance matrix
 * (X - M)T = transposed vector (X - M)
 *
 * NOTE : in the GSL there is no need to transpose a vector before its
 *        multiplication with a matrix
 * NOTE2 : the covariances matrix should be inverted (inv_covs)
 */
double mahalanobis(double *values, double *means, double **inv_covs, int dim)
{
  double dist;
  gsl_vector *gsl_left_vect, *gsl_right_vect;
  gsl_matrix *gsl_mat;

/* allocate memory, initialize vectors and matrix */

  gsl_left_vect = gsl_vector_alloc(dim);
  memcpy(gsl_left_vect->data, values, dim * sizeof(double));
  gsl_right_vect = gsl_vector_alloc(dim);
  memcpy(gsl_right_vect->data, means, dim * sizeof(double));
  gsl_mat = gsl_matrix_alloc(dim, dim);
  memcpy(gsl_mat->data, inv_covs[0], dim * dim * sizeof(double));

/* compute distance */

  gsl_vector_sub(gsl_left_vect, gsl_right_vect);
  gsl_vector_memcpy(gsl_right_vect, gsl_left_vect);
  gsl_blas_dgemv(CblasNoTrans, 1.0, gsl_mat, gsl_right_vect, 0.0, gsl_left_vect);
  gsl_blas_ddot(gsl_left_vect, gsl_right_vect, &dist);

/* free memory */

  gsl_vector_free(gsl_left_vect);
  gsl_vector_free(gsl_right_vect);
  gsl_matrix_free(gsl_mat);

  return dist;
}

/******************************************************************************
 *                                                                            *
 ******************************************************************************/
/**
 * Interchange indx[i] and indx[j]
 */
void swap(int *v, int i, int j)
{
  int temp;

  temp = v[i];
  v[i] = v[j];
  v[j] = temp;
}

/******************************************************************************
 *                                                                            *
 ******************************************************************************/
/**
 * Quick sort in increasing order for double
 */
void dQsort_inc(double *v, int *indx, int left, int right)
{
  int i, last;

/* Do nothing if array contains fewer than two elements */

  if(left >= right)
    return;

/* move partition elements to v[0] */

  swap(indx, left, (left + right) / 2);
  last = left;

/* partition */

  for(i = left + 1; i <= right; i++)
    if(v[indx[i]] < v[indx[left]])
       swap(indx, ++last, i);

/* restore partition element */

  swap(indx, left, last);
  dQsort_inc(v, indx, left, last-1);
  dQsort_inc(v, indx, last+1, right);
  return;
}

/******************************************************************************
 *                                                                            *
 ******************************************************************************/
/**
 * Quick sort in decreasing order for double
 */
void dQsort_dec(double *v, int *indx, int left, int right)
{
  int i, last;

/* Do nothing if array contains fewer than two elements */

  if(left >= right)
    return;

/* move partition elements to v[0] */

  swap(indx, left, (left + right) / 2);
  last = left;

/* partition */

  for(i = left + 1; i <= right; i++)
    if(v[indx[i]] > v[indx[left]])
       swap(indx, ++last, i);

/* restore partition element */

  swap(indx, left, last);
  dQsort_dec(v, indx, left, last-1);
  dQsort_dec(v, indx, last+1, right);
  return;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define PRINT_CUT_OFF 0.01

void print_probas(Data *data)
{
  int i, j, *index = NULL;
  double *tmp_val = NULL, sum;
  Val_data *vdata;

  ALLOC(index, data->num_res, int);
  ALLOC(tmp_val, data->num_res, double);

  for (i = 0; i < data->num_val; i++) {
    vdata = &data->vdata[i];
    for (j = 0; j < data->num_res; j++) {
      index[j] = j;
      tmp_val[j] = vdata->proba[j];
    }
    dQsort_dec(tmp_val, index, 0, data->num_res - 1);

    if (vdata->class == -1) {
      sum = 0.0;
      for (j = 0; j < data->num_res; j++) {
        sum += vdata->proba[j];
      }
      if (sum == 0.0) {
        sum = 1.0;
      }
      printf("%3d: ", i + 1);
      for (j = 0; j < data->num_res; j++) {
        if ((vdata->proba[index[j]] / sum) >= PRINT_CUT_OFF) {
          printf("%c (%.2f) ", AA1[index[j]],
              vdata->proba[index[j]] / sum);
        }
      }
      printf("\n");
    } else {
      printf("# Is class %d\n", vdata->class + 1);
      printf("# Classe");
      for (j = 0; j < data->num_res; j++) {
        printf("%3d", index[j] + 1);
      }
      printf("\n");
    }
  }

  FREE(index);
  FREE(tmp_val);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
double gauss_proba(double val, double mean, double std_dev)
{
  return exp(-0.5 * ((val - mean) / std_dev) * ((val - mean) / std_dev)) /
      (std_dev * sqrt(2.0 * M_PI));
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
double full_perm_indep_proba(Val_data *vdata, Stat_data *sdata)
{
  gsl_permutation *H_perm, *C_perm, *N_perm;
  gsl_combination *H_comb, *C_comb, *N_comb;
  double miss_probaH, miss_probaC, miss_probaN, proba_sum;
  double probaH, probaC, probaN, const_val;
  int *H_tag = NULL, *C_tag = NULL, *N_tag = NULL;
  int i, l, H_size, C_size, N_size, H_shift, C_shift, N_shift;

/* H init */

  if (vdata->nH > 0) {
    H_perm = gsl_permutation_alloc(vdata->nH);
  } else {
    H_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nH > 0) {
    ALLOC(H_tag, sdata->nH, int);
    if (vdata->nH > 0) {
      H_comb = gsl_combination_alloc(sdata->nH, vdata->nH);
    } else {
      H_comb = gsl_combination_alloc(sdata->nH, 1);
    }
  } else {
    if (vdata->nH > 0) {
      H_comb = gsl_combination_alloc(1, vdata->nH);
    } else {
      H_comb = gsl_combination_alloc(1, 1);
    }
  }
  H_size = sdata->nH * sizeof(int);
  H_shift = 0;

/* C init */

  if (vdata->nC > 0) {
    C_perm = gsl_permutation_alloc(vdata->nC);
  } else {
    C_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nC > 0) {
    ALLOC(C_tag, sdata->nC, int);
    if (vdata->nC > 0) {
      C_comb = gsl_combination_alloc(sdata->nC, vdata->nC);
    } else {
      C_comb = gsl_combination_alloc(sdata->nC, 1);
    }
  } else {
    if (vdata->nC > 0) {
      C_comb = gsl_combination_alloc(1, vdata->nC);
    } else {
      C_comb = gsl_combination_alloc(1, 1);
    }
  }
  C_size = sdata->nC * sizeof(int);
  C_shift = sdata->nH;

/* N init */

  if (vdata->nN > 0) {
    N_perm = gsl_permutation_alloc(vdata->nN);
  } else {
    N_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nN > 0) {
    ALLOC(N_tag, sdata->nN, int);
    if (vdata->nN > 0) {
      N_comb = gsl_combination_alloc(sdata->nN, vdata->nN);
    } else {
      N_comb = gsl_combination_alloc(sdata->nN, 1);
    }
  } else {
    if (vdata->nN > 0) {
      N_comb = gsl_combination_alloc(1, vdata->nN);
    } else {
      N_comb = gsl_combination_alloc(1, 1);
    }
  }
  N_size = sdata->nN * sizeof(int);
  N_shift = sdata->nH + sdata->nC;

  const_val = sqrt(2.0 * M_PI);
  proba_sum = 0.0;

/**************************/
/* loop on H combinations */
/**************************/

  gsl_combination_init_first(H_comb);
  do {
    memset(H_tag, 0, H_size);
    for (i = 0; i < vdata->nH; i++) {
      H_tag[H_comb->data[i]] = 1;
    }

    miss_probaH = 1.0;
    for (i = 0; i < sdata->nH; i++) {
      if (H_tag[i] == 0) {
        miss_probaH *= sdata->miss[H_shift + i];
      } else {
        miss_probaH *= 1.0 - sdata->miss[H_shift + i];
      }
    }

  /**************************/
  /* loop on C combinations */
  /**************************/

    gsl_combination_init_first(C_comb);
    do {
      memset(C_tag, 0, C_size);
      for (i = 0; i < vdata->nC; i++) {
        C_tag[C_comb->data[i]] = 1;
      }

      miss_probaC = 1.0;
      for (i = 0; i < sdata->nC; i++) {
        if (C_tag[i] == 0) {
          miss_probaC *= sdata->miss[C_shift + i];
        } else {
          miss_probaC *= (1.0 - sdata->miss[C_shift + i]);
        }
      }

    /**************************/
    /* loop on N combinations */
    /**************************/

      gsl_combination_init_first(N_comb);
      do {
        memset(N_tag, 0, N_size);
        for (i = 0; i < vdata->nN; i++) {
          N_tag[N_comb->data[i]] = 1;
        }

        miss_probaN = 1.0;
        for (i = 0; i < sdata->nN; i++) {
          if (N_tag[i] == 0) {
            miss_probaN *= sdata->miss[N_shift + i];
          } else {
            miss_probaN *= (1.0 - sdata->miss[N_shift + i]);
          }
        }

      /**************************/
      /* loop on H permutations */
      /**************************/

        gsl_permutation_init(H_perm);
        do {

/* present H chemical shifts */

          probaH = 1.0;
          for (i = 0; i < vdata->nH; i++) {
            l = H_comb->data[H_perm->data[i]];
            probaH *=
                exp(-0.5 * ((vdata->H_val[i] - sdata->mean[H_shift + l]) /
                sdata->std_dev[H_shift + l]) * ((vdata->H_val[i] -
                sdata->mean[H_shift + l])) / sdata->std_dev[H_shift + l]) /
                (sdata->std_dev[H_shift + l] * const_val);
          }

        /**************************/
        /* loop on C permutations */
        /**************************/

          gsl_permutation_init(C_perm);
          do {

/* present C chemical shifts */

            probaC = 1.0;
            for (i = 0; i < vdata->nC; i++) {
              l = C_comb->data[C_perm->data[i]];
              probaC *=
                  exp(-0.5 * ((vdata->C_val[i] - sdata->mean[C_shift + l]) /
                  sdata->std_dev[C_shift + l]) * ((vdata->C_val[i] -
                  sdata->mean[C_shift + l])) / sdata->std_dev[C_shift + l]) /
                  (sdata->std_dev[C_shift + l] * const_val);
            }

          /**************************/
          /* loop on N permutations */
          /**************************/

            gsl_permutation_init(N_perm);
            do {

/* present N chemical shifts */

              probaN = 1.0;
              for (i = 0; i < vdata->nN; i++) {
                l = N_comb->data[N_perm->data[i]];
                probaN *=
                    exp(-0.5 * ((vdata->N_val[i] - sdata->mean[N_shift + l]) /
                    sdata->std_dev[N_shift + l]) * ((vdata->N_val[i] -
                    sdata->mean[N_shift + l])) / sdata->std_dev[N_shift + l]) /
                    (sdata->std_dev[N_shift + l] * const_val);
              }

              proba_sum += miss_probaH * probaH * miss_probaC * probaC *
                  miss_probaN * probaN;

            } while (gsl_permutation_next(N_perm) == GSL_SUCCESS);
          } while (gsl_permutation_next(C_perm) == GSL_SUCCESS);
        } while (gsl_permutation_next(H_perm) == GSL_SUCCESS);
      } while (gsl_combination_next(N_comb) == GSL_SUCCESS);
    } while (gsl_combination_next(C_comb) == GSL_SUCCESS);
  } while (gsl_combination_next(H_comb) == GSL_SUCCESS);

  gsl_permutation_free(H_perm);
  gsl_combination_free(H_comb);
  gsl_permutation_free(C_perm);
  gsl_combination_free(C_comb);
  gsl_permutation_free(N_perm);
  gsl_combination_free(N_comb);
  if (H_tag != NULL) {
    FREE(H_tag);
  }
  if (C_tag != NULL) {
    FREE(C_tag);
  }
  if (N_tag != NULL) {
    FREE(N_tag);
  }

  return proba_sum;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate independant probabilities using full permutations
 */
void calculate_full_perm_indep_probas(Data *data)
{
  int class, val;
  Stat_data *sdata;
  Val_data *vdata;

  for (val = 0; val < data->num_val; val++) {
    vdata = &data->vdata[val];
    for (class = 0; class < data->num_res; class++) {
      sdata = &data->sdata[class];
      if (vdata->nH <= sdata->nH && vdata->nC <= sdata->nC &&
          vdata->nN <= sdata->nN) {
        vdata->proba[class] = full_perm_indep_proba(vdata, sdata);
      } else {
        vdata->proba[class] = 0.0;
      }
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate dependant probabilities using multivariate normal distributions
 *
 * To be able to compute very small values, it is done like this :
 * Proba = exp(A - B - C)
 * with :
 * A = -1/2 Mahalanobis_distance
 * B = (log(2 Pi) dim)/2 (dim = number of vector dimension)
 * C = cov_det = 1/2 log(det(sigma))
 * (computed with a special function that uses logs)
 */
double full_perm_dep_proba(Val_data *vdata, Stat_data *sdata)
{
  gsl_permutation *H_perm, *C_perm, *N_perm;
  gsl_combination *H_comb, *C_comb, *N_comb;
  double miss_probaH, miss_probaC, miss_probaN, proba_sum;
  double A, B, mahalanobis_dist, **cov, **inv_cov, log_det;
  double *values, *means;
  int *H_tag = NULL, *C_tag = NULL, *N_tag = NULL;
  int i, j, k, l, H_size, C_size, N_size;
  int H_sshift, C_sshift, N_sshift, H_vshift, C_vshift, N_vshift;

/* Initialize mahalanobis stuff */

  ALLOC(values, vdata->n_total, double);
  ALLOC(means, vdata->n_total, double);
  ALLOC2(cov, vdata->n_total, vdata->n_total, double);
  ALLOC2(inv_cov, vdata->n_total, vdata->n_total, double);
  B = log(2.0 * M_PI) * (double) vdata->n_total / 2.0;

/* H init */

  if (vdata->nH > 0) {
    H_perm = gsl_permutation_alloc(vdata->nH);
  } else {
    H_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nH > 0) {
    ALLOC(H_tag, sdata->nH, int);
    if (vdata->nH > 0) {
      H_comb = gsl_combination_alloc(sdata->nH, vdata->nH);
    } else {
      H_comb = gsl_combination_alloc(sdata->nH, 1);
    }
  } else {
    if (vdata->nH > 0) {
      H_comb = gsl_combination_alloc(1, vdata->nH);
    } else {
      H_comb = gsl_combination_alloc(1, 1);
    }
  }
  H_size = sdata->nH * sizeof(int);
  H_sshift = 0;
  H_vshift = 0;

/* C init */

  if (vdata->nC > 0) {
    C_perm = gsl_permutation_alloc(vdata->nC);
  } else {
    C_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nC > 0) {
    ALLOC(C_tag, sdata->nC, int);
    if (vdata->nC > 0) {
      C_comb = gsl_combination_alloc(sdata->nC, vdata->nC);
    } else {
      C_comb = gsl_combination_alloc(sdata->nC, 1);
    }
  } else {
    if (vdata->nC > 0) {
      C_comb = gsl_combination_alloc(1, vdata->nC);
    } else {
      C_comb = gsl_combination_alloc(1, 1);
    }
  }
  C_size = sdata->nC * sizeof(int);
  C_sshift = sdata->nH;
  C_vshift = vdata->nH;

/* N init */

  if (vdata->nN > 0) {
    N_perm = gsl_permutation_alloc(vdata->nN);
  } else {
    N_perm = gsl_permutation_alloc(1);
  }
  if (sdata->nN > 0) {
    ALLOC(N_tag, sdata->nN, int);
    if (vdata->nN > 0) {
      N_comb = gsl_combination_alloc(sdata->nN, vdata->nN);
    } else {
      N_comb = gsl_combination_alloc(sdata->nN, 1);
    }
  } else {
    if (vdata->nN > 0) {
      N_comb = gsl_combination_alloc(1, vdata->nN);
    } else {
      N_comb = gsl_combination_alloc(1, 1);
    }
  }
  N_size = sdata->nN * sizeof(int);
  N_sshift = sdata->nH + sdata->nC;
  N_vshift = vdata->nH + vdata->nC;

  proba_sum = 0.0;

/**************************/
/* loop on H combinations */
/**************************/

  gsl_combination_init_first(H_comb);
  do {

    memset(H_tag, 0, H_size);
    for (i = 0; i < vdata->nH; i++) {
      k = H_comb->data[i];
      H_tag[k] = 1;
      k += H_sshift;
      l = H_vshift + i;
      means[l] = sdata->mean[k];
      cov[l][l] = sdata->cov[k][k];
      for (j = i + 1; j < vdata->nH; j++) {
        cov[l][H_vshift + j] = cov[H_vshift + j][l] =
            sdata->cov[k][H_sshift + H_comb->data[j]];
      }
    }

    miss_probaH = 1.0;
    for (i = 0; i < sdata->nH; i++) {
      if (H_tag[i] == 0) {
        miss_probaH *= sdata->miss[H_sshift + i];
      } else {
        miss_probaH *= (1.0 - sdata->miss[H_sshift + i]);
      }
    }

  /**************************/
  /* loop on C combinations */
  /**************************/

    gsl_combination_init_first(C_comb);
    do {

      memset(C_tag, 0, C_size);
      for (i = 0; i < vdata->nC; i++) {
        k = C_comb->data[i];
        C_tag[k] = 1;
        k += C_sshift;
        l = C_vshift + i;
        means[l] = sdata->mean[k];
        cov[l][l] = sdata->cov[k][k];
        for (j = i + 1; j < vdata->nC; j++) {
          cov[l][C_vshift + j] = cov[C_vshift + j][l] =
              sdata->cov[k][C_sshift + C_comb->data[j]];
        }
        for (j = 0; j < vdata->nH; j++) {
          cov[l][H_vshift + j] = cov[H_vshift + j][l] =
              sdata->cov[k][H_sshift + H_comb->data[j]];
        }
      }

      miss_probaC = 1.0;
      for (i = 0; i < sdata->nC; i++) {
        if (C_tag[i] == 0) {
          miss_probaC *= sdata->miss[C_sshift + i];
        } else {
          miss_probaC *= (1.0 - sdata->miss[C_sshift + i]);
        }
      }

    /**************************/
    /* loop on N combinations */
    /**************************/

      gsl_combination_init_first(N_comb);
      do {
        memset(N_tag, 0, N_size);
        for (i = 0; i < vdata->nN; i++) {
          k = N_comb->data[i];
          N_tag[k] = 1;
          k += N_sshift;
          l = N_vshift + i;
          means[l] = sdata->mean[k];
          cov[l][l] = sdata->cov[k][k];
          for (j = i + 1; j < vdata->nN; j++) {
            cov[l][N_vshift + j] = cov[N_vshift + j][l] =
                sdata->cov[k][N_sshift + N_comb->data[j]];
          }
          for (j = 0; j < vdata->nC; j++) {
            cov[l][C_vshift + j] = cov[C_vshift + j][l] =
                sdata->cov[k][C_sshift + C_comb->data[j]];
          }
          for (j = 0; j < vdata->nH; j++) {
            cov[l][H_vshift + j] = cov[H_vshift + j][l] =
                sdata->cov[k][H_sshift + H_comb->data[j]];
          }
        }

        miss_probaN = 1.0;
        for (i = 0; i < sdata->nN; i++) {
          if (N_tag[i] == 0) {
            miss_probaN *= sdata->miss[N_sshift + i];
          } else {
            miss_probaN *= (1.0 - sdata->miss[N_sshift + i]);
          }
        }
/*
        for (i = 0; i < vdata->n_total; i++) {
          for (j = 0; j < vdata->n_total; j++) {
            if (i != j) {
              cov[i][j] = 0.0;
            }
            
          }
        }
*/
        log_det = lu_matrix_log_determinant(cov, vdata->n_total) / 2.0;
        lu_decomposition_mat_inv(cov, vdata->n_total, inv_cov);

      /**************************/
      /* loop on H permutations */
      /**************************/

        gsl_permutation_init(H_perm);
        do {

/* set H part of values */

          for (i = 0; i < vdata->nH; i++) {
            values[H_vshift + i] = vdata->H_val[H_perm->data[i]];
          }

        /**************************/
        /* loop on C permutations */
        /**************************/

          gsl_permutation_init(C_perm);
          do {

/* set C part of values */

            for (i = 0; i < vdata->nC; i++) {
              values[C_vshift + i] = vdata->C_val[C_perm->data[i]];
            }

          /**************************/
          /* loop on N permutations */
          /**************************/

            gsl_permutation_init(N_perm);
            do {

/* set N part of values */

              for (i = 0; i < vdata->nN; i++) {
                values[N_vshift + i] = vdata->N_val[N_perm->data[i]];
              }

/* calculate proba */

              mahalanobis_dist = mahalanobis(values, means, inv_cov,
                  vdata->n_total);

              if (mahalanobis_dist < 0.0) {
                log_det = -log_det;
                mahalanobis_dist = -mahalanobis_dist;
              }
/*
              if (mahalanobis_dist < 0.0) {
                fprintf(stderr, "Error: negative Mahalanobis distance !\n");
                fprintf(stderr, "Values:\n");
                for (i = 0; i < vdata->n_total; i++) {
                  fprintf(stderr, "%10.4f", values[i]);
                }
                fprintf(stderr, "\n");
                fprintf(stderr, "Means:\n");
                for (i = 0; i < vdata->n_total; i++) {
                  fprintf(stderr, "%10.4f", means[i]);
                }
                fprintf(stderr, "\n");
                fprintf(stderr, "Covariances:\n");
                for (i = 0; i < vdata->n_total; i++) {
                  for (j = 0; j < vdata->n_total; j++) {
                    fprintf(stderr, "%10.4f", cov[i][j]);
                  }
                  fprintf(stderr, "\n");
                }
                fprintf(stderr, "log det = %.6g\n", log_det);
                fprintf(stderr, "mahalanobis distance = %.6g\n",
                    mahalanobis_dist);
                exit(1);
              }
*/
              A = -mahalanobis_dist / 2.0;
              proba_sum += miss_probaH * miss_probaC * miss_probaN *
                  exp(A - B - log_det);
/*
              printf("miss_probaH = %.4g\nmiss_probaC = %.4g\n"
                  "miss_probaN = %.4g\n", miss_probaH, miss_probaC, miss_probaN);
              printf("proba_sum = %.4g\n\n", proba_sum);
*/
            } while (gsl_permutation_next(N_perm) == GSL_SUCCESS);
          } while (gsl_permutation_next(C_perm) == GSL_SUCCESS);
        } while (gsl_permutation_next(H_perm) == GSL_SUCCESS);
      } while (gsl_combination_next(N_comb) == GSL_SUCCESS);
    } while (gsl_combination_next(C_comb) == GSL_SUCCESS);
  } while (gsl_combination_next(H_comb) == GSL_SUCCESS);

  gsl_permutation_free(H_perm);
  gsl_combination_free(H_comb);
  gsl_permutation_free(C_perm);
  gsl_combination_free(C_comb);
  gsl_permutation_free(N_perm);
  gsl_combination_free(N_comb);
  if (H_tag != NULL) {
    FREE(H_tag);
  }
  if (C_tag != NULL) {
    FREE(C_tag);
  }
  if (N_tag != NULL) {
    FREE(N_tag);
  }
  FREE(values);
  FREE(means);
  FREE2(cov);
  FREE2(inv_cov);

  return proba_sum;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate multivariate probabilities using full permutations
 */
void calculate_full_perm_dep_probas(Data *data)
{
  int class, val;
  Stat_data *sdata;
  Val_data *vdata;

  for (val = 0; val < data->num_val; val++) {
    vdata = &data->vdata[val];
    for (class = 0; class < data->num_res; class++) {
      sdata = &data->sdata[class];
      if (vdata->nH <= sdata->nH && vdata->nC <= sdata->nC &&
          vdata->nN <= sdata->nN) {
        vdata->proba[class] = full_perm_dep_proba(vdata, sdata);
      } else {
        vdata->proba[class] = 0.0;
      }
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * 1) Set comb and comb_max used for calculating combinations and
 *    set perm and perm_max used to calculate permutations
 * 2) check that each observation can belong to a distribution
 * 3) check that each distribution has not more than one exclusive observation
 */
int init_enum_data(Enum_data *enumd, double *mean, double *std_dev,
    int sn, int shift, int vn, double *val)
{
  int i, j, k, l, flag;

  for (i = 0; i < vn; i++) {
    for (j = 0; j < sn; j++) {

#ifdef DEBUG
      printf("proba v (%6.2f): %d, s (%6.2f): %d = %g\n", val[i], i,
          mean[shift + j], j, gauss_proba(val[i], mean[shift + j],
          std_dev[shift + j]));
#endif

      if (gauss_proba(val[i], mean[shift + j], std_dev[shift + j]) >
          PROBA_CUT_OFF) {
        enumd->val[i][enumd->max_val[i]] = j;
        enumd->max_val[i]++;
      }
    }
    if (enumd->max_val[i] == 0) {
      return 1;
    }
  }

#ifdef DEBUG
  printf("Selected enumerations:\n");
  for (i = 0; i < vn; i++) {
    printf("%d(%d): ", i, enumd->max_val[i]);
    for (j = 0; j < enumd->max_val[i]; j++) {
      printf("%d ", enumd->val[i][j]);
    }
    printf("\n");
  }
  fflush(stdout);
#endif

/* enumerations can be simplified */

  do {
    flag = 0;

/* remove exclusive attributions from other selections */

    for (i = 0; i < vn; i++) {
      if (enumd->max_val[i] == 1) {
        for (j = 0; j < vn; j++) {
          if (j != i) {
            for (k = 0; k < enumd->max_val[j]; k++) {
              if (enumd->val[i][0] == enumd->val[j][k]) {
                if (enumd->max_val[j] == 1) {
                  return 2;
                }
                for (l = k + 1; l < enumd->max_val[j]; l++) {
                  enumd->val[j][l - 1] = enumd->val[j][l];
                }
                enumd->max_val[j]--;
                flag = 1;
              }
            }
          }
        }
      }
    }
  } while (flag != 0);

#ifdef DEBUG
  printf("Selected enumerations after simplification:\n");
  for (i = 0; i < vn; i++) {
    printf("%d(%d): ", i, enumd->max_val[i]);
    for (j = 0; j < enumd->max_val[i]; j++) {
      printf("%d ", enumd->val[i][j]);
    }
    printf("\n");
  }
  fflush(stdout);
#endif

  return 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void alloc_enum_data(Enum_data *enumd, int i, int j)
{
  CALLOC2(enumd->val, i, j, int);
  CALLOC(enumd->max_val, i, int);
  CALLOC(enumd->pos, i, int);
  ALLOC(enumd->tag, j, int);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void free_enum_data(Enum_data *enumd)
{
  FREE2(enumd->val);
  FREE(enumd->max_val);
  FREE(enumd->pos);
  FREE(enumd->tag);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_enum(Enum_data *enumd, int vn, int sn)
{
  int i;

  printf(" + pos:\n");
  for (i = 0; i < vn; i++) {
    printf("%4d", enumd->pos[i]);
  }
  printf("\n");
  printf(" + tag:\n");
  for (i = 0; i < sn; i++) {
    printf("%4d", enumd->tag[i]);
  }
  printf("\n");
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * THE nightmare function !
 * Select the next enumeration (permutation of combination)
 */
int select_enum_next(Enum_data *enumd, int vn)
{
  int cur_pos, ntag;

  if (vn == 0) {
    return 1;
  }

  vn--;
  cur_pos = ntag = vn;

/*
 * The following test is necessary while this function is used by
 * the 'select_enum_init_first' function
 */
  if (enumd->pos[cur_pos] >= 0) {
    enumd->tag[enumd->val[cur_pos][enumd->pos[cur_pos]]] = -1;
  }
  ntag--;
  do {
    enumd->pos[cur_pos]++;
    if (enumd->pos[cur_pos] == enumd->max_val[cur_pos]) {
      enumd->pos[cur_pos] = -1;
      cur_pos--;
      if (cur_pos >= 0) {

/*
 * The following test is necessary while this function is used by
 * the 'select_enum_init_first' function
 */
        if (enumd->pos[cur_pos] >= 0) {
          enumd->tag[enumd->val[cur_pos][enumd->pos[cur_pos]]] = -1;
        }
        ntag--;
      } else {
        return 1;
      }
    } else {
      if (enumd->tag[enumd->val[cur_pos][enumd->pos[cur_pos]]] == -1) {
        enumd->tag[enumd->val[cur_pos][enumd->pos[cur_pos]]] = cur_pos;
        ntag++;
        if (cur_pos != vn) {
          cur_pos++;
        }
      }
    }
  } while (ntag != vn);

  if (cur_pos < 0) {
    return 1;
  }

  return 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int select_enum_init_first(Enum_data *enumd, int vn, int sn)
{
  int i;

  for (i = 0; i < sn; i++) {
    enumd->tag[i] = -1;
  }

  if (vn == 0) {
    return 0;
  }

  enumd->pos[0] = -1;
  for (i = 1; i < vn; i++) {
    enumd->pos[i] = enumd->max_val[i] - 1;
  }

  if (select_enum_next(enumd, vn) != 0) {
    return 1;
  }
  return 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
double select_perm_indep_proba(Val_data *vdata, Stat_data *sdata)
{
  Enum_data H_enum, C_enum, N_enum;
  int i, H_shift, C_shift, N_shift;
  double probaH, probaC, probaN, proba_sum;

/* Init H */

  H_shift = 0;
  if (vdata->nH == 0) {
    if (sdata->nH == 0) {
      alloc_enum_data(&H_enum, 1, 1);
    } else {
      alloc_enum_data(&H_enum, 1, sdata->nH);
    }
  } else {
    if (sdata->nH == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&H_enum, vdata->nH, sdata->nH);
    }
  }
  if (init_enum_data(&H_enum, sdata->mean, sdata->std_dev, sdata->nH,
      H_shift, vdata->nH, vdata->H_val) != 0) {
    free_enum_data(&H_enum);
    return 0.0;
  }

/* Init C */

  C_shift = sdata->nH;
  if (vdata->nC == 0) {
    if (sdata->nC == 0) {
      alloc_enum_data(&C_enum, 1, 1);
    } else {
      alloc_enum_data(&C_enum, 1, sdata->nC);
    }
  } else {
    if (sdata->nC == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&C_enum, vdata->nC, sdata->nC);
    }
  }
  if (init_enum_data(&C_enum, sdata->mean, sdata->std_dev, sdata->nC,
      C_shift, vdata->nC, vdata->C_val) != 0) {
    free_enum_data(&C_enum);
    return 0.0;
  }

/* Init N */

  N_shift = sdata->nH + sdata->nC;
  if (vdata->nN == 0) {
    if (sdata->nN == 0) {
      alloc_enum_data(&N_enum, 1, 1);
    } else {
      alloc_enum_data(&N_enum, 1, sdata->nN);
    }
  } else {
    if (sdata->nN == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&N_enum, vdata->nN, sdata->nN);
    }
  }
  if (init_enum_data(&N_enum, sdata->mean, sdata->std_dev, sdata->nN,
      N_shift, vdata->nN, vdata->N_val) != 0) {
    free_enum_data(&N_enum);
    return 0.0;
  }

  proba_sum = 0.0;

/**************************/
/* loop on H enumerations */
/**************************/

  if (select_enum_init_first(&H_enum, vdata->nH, sdata->nH) != 0) {
    return 0.0;
  }
  do {

    /*print_enum(&H_enum, vdata->nH, sdata->nH);*/
    probaH = 1.0;
    for (i = 0; i < sdata->nH; i++) {
      if (H_enum.tag[i] != -1) {
        probaH *= (1.0 - sdata->miss[H_shift + i]) *
            gauss_proba(vdata->H_val[H_enum.tag[i]],
            sdata->mean[H_shift + i], sdata->std_dev[H_shift + i]);
      } else {
        probaH *= sdata->miss[H_shift + i];
      }
    }

  /**************************/
  /* loop on C enumerations */
  /**************************/

    if (select_enum_init_first(&C_enum, vdata->nC, sdata->nC) != 0) {
      return 0.0;
    }
    do {

      /*print_enum(&C_enum, vdata->nC, sdata->nC);*/
      probaC = 1.0;
      for (i = 0; i < sdata->nC; i++) {
        if (C_enum.tag[i] != -1) {
          probaC *= (1.0 - sdata->miss[C_shift + i]) *
              gauss_proba(vdata->C_val[C_enum.tag[i]],
              sdata->mean[C_shift + i], sdata->std_dev[C_shift + i]);
        } else {
          probaC *= sdata->miss[C_shift + i];
        }
      }

    /**************************/
    /* loop on N enumerations */
    /**************************/

      if (select_enum_init_first(&N_enum, vdata->nN, sdata->nN) != 0) {
        return 0.0;
      }
      do {

        /*print_enum(&N_enum, vdata->nN, sdata->nN);*/
        probaN = 1.0;
        for (i = 0; i < sdata->nN; i++) {
          if (N_enum.tag[i] != -1) {
            probaN *= (1.0 - sdata->miss[N_shift + i]) *
                gauss_proba(vdata->N_val[N_enum.tag[i]],
                sdata->mean[N_shift + i], sdata->std_dev[N_shift + i]);
          } else {
            probaN *= sdata->miss[N_shift + i];
          }
        }

        proba_sum += probaH * probaC * probaN;
/*
        printf("proba_sum = %.4g (H: %.4g, C: %.4g, N: %.4g)\n", proba_sum,
            probaH, probaC, probaN);
*/
      } while (select_enum_next(&N_enum, vdata->nN) == 0);
    } while (select_enum_next(&C_enum, vdata->nC) == 0);
  } while (select_enum_next(&H_enum, vdata->nH) == 0);

/* free memory */

  free_enum_data(&H_enum);
  free_enum_data(&C_enum);
  free_enum_data(&N_enum);

  return proba_sum;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate independant probabilities using only some selected permutations
 */
void calculate_select_perm_indep_probas(Data *data)
{
  int class, val;
  Stat_data *sdata;
  Val_data *vdata;

  for (val = 0; val < data->num_val; val++) {
    vdata = &data->vdata[val];
    for (class = 0; class < data->num_res; class++) {
      sdata = &data->sdata[class];
      if (vdata->nH <= sdata->nH && vdata->nC <= sdata->nC &&
          vdata->nN <= sdata->nN) {
        vdata->proba[class] = select_perm_indep_proba(vdata, sdata);
      } else {
        vdata->proba[class] = 0.0;
      }
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
double select_perm_dep_proba(Val_data *vdata, Stat_data *sdata)
{
  Enum_data H_enum, C_enum, N_enum;
  int i, j;
  int H_sshift, C_sshift, N_sshift, H_vshift, C_vshift, N_vshift;
  double miss_probaH, miss_probaC, miss_probaN;
  double A, B, mahalanobis_dist, **cov, **inv_cov, log_det;
  double *values, *means, proba_sum;

/* Initialize mahalanobis stuff */

  ALLOC(values, vdata->n_total, double);
  ALLOC(means, vdata->n_total, double);
  ALLOC2(cov, vdata->n_total, vdata->n_total, double);
  ALLOC2(inv_cov, vdata->n_total, vdata->n_total, double);
  B = log(2.0 * M_PI) * (double) vdata->n_total / 2.0;

/* Init H */

  H_sshift = 0;
  H_vshift = 0;
  if (vdata->nH == 0) {
    if (sdata->nH == 0) {
      alloc_enum_data(&H_enum, 1, 1);
    } else {
      alloc_enum_data(&H_enum, 1, sdata->nH);
    }
  } else {
    if (sdata->nH == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&H_enum, vdata->nH, sdata->nH);
    }
  }
  if (init_enum_data(&H_enum, sdata->mean, sdata->std_dev, sdata->nH,
      H_sshift, vdata->nH, vdata->H_val) != 0) {
    free_enum_data(&H_enum);
    return 0.0;
  }

/* Init C */

  C_sshift = sdata->nH;
  C_vshift = vdata->nH;
  if (vdata->nC == 0) {
    if (sdata->nC == 0) {
      alloc_enum_data(&C_enum, 1, 1);
    } else {
      alloc_enum_data(&C_enum, 1, sdata->nC);
    }
  } else {
    if (sdata->nC == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&C_enum, vdata->nC, sdata->nC);
    }
  }
  if (init_enum_data(&C_enum, sdata->mean, sdata->std_dev, sdata->nC,
      C_sshift, vdata->nC, vdata->C_val) != 0) {
    free_enum_data(&C_enum);
    return 0.0;
  }

/* Init N */

  N_sshift = sdata->nH + sdata->nC;
  N_vshift = vdata->nH + vdata->nC;
  if (vdata->nN == 0) {
    if (sdata->nN == 0) {
      alloc_enum_data(&N_enum, 1, 1);
    } else {
      alloc_enum_data(&N_enum, 1, sdata->nN);
    }
  } else {
    if (sdata->nN == 0) {
      return 0.0;
    } else {
      alloc_enum_data(&N_enum, vdata->nN, sdata->nN);
    }
  }
  if (init_enum_data(&N_enum, sdata->mean, sdata->std_dev, sdata->nN,
      N_sshift, vdata->nN, vdata->N_val) != 0) {
    free_enum_data(&N_enum);
    return 0.0;
  }

  proba_sum = 0.0;

/**************************/
/* loop on H enumerations */
/**************************/

  if (select_enum_init_first(&H_enum, vdata->nH, sdata->nH) != 0) {
    return 0.0;
  }
  do {

    /*print_enum(&H_enum, vdata->nH, sdata->nH);*/
    miss_probaH = 1.0;
    for (i = 0; i < sdata->nH; i++) {
      if (H_enum.tag[i] == -1) {
        miss_probaH *= sdata->miss[H_sshift + i];
      } else {
        miss_probaH *= (1.0 - sdata->miss[H_sshift + i]);
      }
    }

/* set values, means and cov */

    for (i = 0; i < vdata->nH; i++) {
      values[H_vshift + i] = vdata->H_val[i];
      means[H_vshift + i] =
          sdata->mean[H_sshift + H_enum.val[i][H_enum.pos[i]]];
      cov[H_vshift + i][H_vshift + i] =
          sdata->cov[H_sshift + H_enum.val[i][H_enum.pos[i]]]
          [H_sshift + H_enum.val[i][H_enum.pos[i]]];
      for (j = i + 1; j < vdata->nH; j++) {
        cov[H_vshift + i][H_vshift + j] = cov[H_vshift + j][H_vshift + i] =
            sdata->cov[H_sshift + H_enum.val[i][H_enum.pos[i]]]
            [H_sshift + H_enum.val[j][H_enum.pos[j]]];
      }
    }

  /**************************/
  /* loop on C enumerations */
  /**************************/

    if (select_enum_init_first(&C_enum, vdata->nC, sdata->nC) != 0) {
      return 0.0;
    }
    do {

      /*print_enum(&C_enum, vdata->nC, sdata->nC);*/
      miss_probaC = 1.0;
      for (i = 0; i < sdata->nC; i++) {
        if (C_enum.tag[i] == -1) {
          miss_probaC *= sdata->miss[C_sshift + i];
        } else {
          miss_probaC *= (1.0 - sdata->miss[C_sshift + i]);
        }
      }

/* set values, means and cov */

      for (i = 0; i < vdata->nC; i++) {
        values[C_vshift + i] = vdata->C_val[i];
        means[C_vshift + i] =
            sdata->mean[C_sshift + C_enum.val[i][C_enum.pos[i]]];
        cov[C_vshift + i][C_vshift + i] =
            sdata->cov[C_sshift + C_enum.val[i][C_enum.pos[i]]]
            [C_sshift + C_enum.val[i][C_enum.pos[i]]];
        for (j = i + 1; j < vdata->nC; j++) {
          cov[C_vshift + i][C_vshift + j] = cov[C_vshift + j][C_vshift + i] =
              sdata->cov[C_sshift + C_enum.val[i][C_enum.pos[i]]]
              [C_sshift + C_enum.val[j][C_enum.pos[j]]];
        }
        for (j = 0; j < vdata->nH; j++) {
          cov[C_vshift + i][H_vshift + j] = cov[H_vshift + j][C_vshift + i] =
              sdata->cov[C_sshift + C_enum.val[i][C_enum.pos[i]]]
              [H_sshift + H_enum.val[j][H_enum.pos[j]]];
        }
      }

    /**************************/
    /* loop on N enumerations */
    /**************************/

      if (select_enum_init_first(&N_enum, vdata->nN, sdata->nN) != 0) {
        return 0.0;
      }
      do {

        /*print_enum(&N_enum, vdata->nN, sdata->nN);*/
        miss_probaN = 1.0;
        for (i = 0; i < sdata->nN; i++) {
          if (N_enum.tag[i] == -1) {
            miss_probaN *= sdata->miss[N_sshift + i];
          } else {
            miss_probaN *= (1.0 - sdata->miss[N_sshift + i]);
          }
        }

/* set values, means and cov */

        for (i = 0; i < vdata->nN; i++) {
          values[N_vshift + i] = vdata->N_val[i];
          means[N_vshift + i] =
              sdata->mean[N_sshift + N_enum.val[i][N_enum.pos[i]]];
          cov[N_vshift + i][N_vshift + i] =
              sdata->cov[N_sshift + N_enum.val[i][N_enum.pos[i]]]
              [N_sshift + N_enum.val[i][N_enum.pos[i]]];
          for (j = i + 1; j < vdata->nN; j++) {
            cov[N_vshift + i][N_vshift + j] = cov[N_vshift + j][N_vshift + i] =
                sdata->cov[N_sshift + N_enum.val[i][N_enum.pos[i]]]
                [N_sshift + N_enum.val[j][N_enum.pos[j]]];
          }
          for (j = 0; j < vdata->nC; j++) {
            cov[N_vshift + i][C_vshift + j] = cov[C_vshift + j][N_vshift + i] =
                sdata->cov[N_sshift + N_enum.val[i][N_enum.pos[i]]]
                [C_sshift + C_enum.val[j][C_enum.pos[j]]];
          }
          for (j = 0; j < vdata->nH; j++) {
            cov[N_vshift + i][H_vshift + j] = cov[H_vshift + j][N_vshift + i] =
                sdata->cov[N_sshift + N_enum.val[i][N_enum.pos[i]]]
                [H_sshift + H_enum.val[j][H_enum.pos[j]]];
          }
        }
/*
        for (i = 0; i < vdata->n_total; i++) {
          for (j = 0; j < vdata->n_total; j++) {
            if (i != j) {
              cov[i][j] = 0.0;
            }
            
          }
        }
*/
        log_det = lu_matrix_log_determinant(cov, vdata->n_total) / 2.0;
        lu_decomposition_mat_inv(cov, vdata->n_total, inv_cov);

        mahalanobis_dist = mahalanobis(values, means, inv_cov, vdata->n_total);

        if (mahalanobis_dist < 0.0) {
          log_det = -log_det;
          mahalanobis_dist = -mahalanobis_dist;
        }
/*
        if (mahalanobis_dist < 0.0) {
          fprintf(stderr, "Error: negative Mahalanobis distance !\n");
          fprintf(stderr, "Values:\n");
          for (i = 0; i < vdata->n_total; i++) {
            fprintf(stderr, "%10.4f", values[i]);
          }
          fprintf(stderr, "\n");
          fprintf(stderr, "Means:\n");
          for (i = 0; i < vdata->n_total; i++) {
            fprintf(stderr, "%10.4f", means[i]);
          }
          fprintf(stderr, "\n");
          fprintf(stderr, "Covariances:\n");
          for (i = 0; i < vdata->n_total; i++) {
            for (j = 0; j < vdata->n_total; j++) {
              fprintf(stderr, "%10.4f", cov[i][j]);
            }
            fprintf(stderr, "\n");
          }
          fprintf(stderr, "log det = %.6g\n", log_det);
          fprintf(stderr, "mahalanobis distance = %.6g\n", mahalanobis_dist);
          exit(1);
        }
*/
        A = -mahalanobis_dist / 2.0;
        proba_sum += miss_probaH * miss_probaC * miss_probaN *
            exp(A - B - log_det);
/*
        printf("proba_sum = %.4g (H: %.4g, C: %.4g, N: %.4g)\n", proba_sum,
            probaH, probaC, probaN);
*/
      } while (select_enum_next(&N_enum, vdata->nN) == 0);
    } while (select_enum_next(&C_enum, vdata->nC) == 0);
  } while (select_enum_next(&H_enum, vdata->nH) == 0);

/* free memory */

  free_enum_data(&H_enum);
  free_enum_data(&C_enum);
  free_enum_data(&N_enum);

  return proba_sum;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
/**
 * Calculate multivariate probabilities using only some selected permutations
 */
void calculate_select_perm_dep_probas(Data *data)
{
  int class, val;
  Stat_data *sdata;
  Val_data *vdata;

  for (val = 0; val < data->num_val; val++) {
    vdata = &data->vdata[val];
    for (class = 0; class < data->num_res; class++) {
      sdata = &data->sdata[class];
      if (vdata->nH <= sdata->nH && vdata->nC <= sdata->nC &&
          vdata->nN <= sdata->nN) {
        vdata->proba[class] = select_perm_dep_proba(vdata, sdata);
      } else {
        vdata->proba[class] = 0.0;
      }
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  Parameters params;
  Data data;

  init_parameters(&params);
  process_command_line(argc, argv, &params);

  read_cs_to_analyse(&params, &data);
  read_stats(&params, &data);
  read_input_values(&data);

/* full permutation independant probabilities */

  if (params.full_perm_indep == 1) {
    calculate_full_perm_indep_probas(&data);
    if (params.verbosity >= 3) {
      printf("# Full permutations independant probabilities\n");
    }
    print_probas(&data);
  }

/* full permutation dependant probabilities (using covariances) */

  if (params.full_perm_dep == 1) {
    calculate_full_perm_dep_probas(&data);
    if (params.verbosity >= 3) {
      printf("# Full permutations MVND probabilities\n");
    }
    print_probas(&data);
  }

/* most probable permutations independant probabilities */

  if (params.select_perm_indep == 1) {
    calculate_select_perm_indep_probas(&data);
    if (params.verbosity >= 3) {
      printf("# Selected permutations independant probabilities\n");
    }
    print_probas(&data);
  }

/* most probable permutations dependant probabilities (using covariances) */

  if (params.select_perm_dep == 1) {
    calculate_select_perm_dep_probas(&data);
    if (params.verbosity >= 3) {
      printf("# Selected permutations MVND probabilities\n");
    }
    print_probas(&data);
  }

  return 0;
}
