/*
 * svmtest_results_to_graph.c
 * This program generates some useful figures to analyse the rescue2 results
 *
 * Copyright (C) 2003 Antoine MARIN
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License, version 2.1 published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact the author via mail or email at:
 * Antoine Marin (Antoine.Marin@ibpc.fr)
 * CNRS UPR 9080
 * Laboratoire de Biochimie Thorique
 * 13, rue Pierre et Marie Curie
 * 75005 Paris - France
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h> /* for basename */

#include "alloc_utils.h"

#define VERSION "1.0"
#define MAX_BUF 1024

char *AA1 = "ARDNCEQGHILKMFPSTWYV";

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
typedef struct parameters {
  int dump_matrix;
  char dump_file_name[MAX_BUF];
} Parameters;

typedef struct data {
  int num_class;
  int *total_class;
  int *total_obs;
  int total;
  int **mat;
} Data;

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void init_parameters(Parameters *params)
{
  params->dump_matrix = 0;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void print_usage(FILE *fp, char **argv)
{
  fprintf(fp, "\nUsage: %s [options] < SVMTest_mod_result_file\n\n",
      basename(argv[0]));
  fprintf(fp, "  Options:\n\n");
  fprintf(fp, "    -o file_name  : output matrix file.\n\n");
  fprintf(fp, "    -v, --version : print version only (current version = %s).\n\n",
      VERSION);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void process_command_line(int argc, char **argv, Parameters *params)
{
  int i;

  for (i = 1; i < argc; i++) {

    if (strcmp(argv[i], "-o") == 0)  {

/* -o : output file */

      params->dump_matrix=1;
      i++;
      strcpy(params->dump_file_name, argv[i]);
    } else if (strcmp(argv[i], "-h") == 0 || strcmp(argv[i], "--help") == 0) {

/* -h, --help */

      print_usage(stdout, argv);
      exit(0);
    } else if (strcmp(argv[i], "-v") == 0 ||
      strcmp(argv[i], "--version") == 0) {

/* -v, --version : print version and exit */

      fprintf(stdout, "This is %s release number %s\n", basename(argv[0]),
          VERSION);
      exit(0);
    } else {

/* Unknown tag */

      fprintf(stderr, "Error: unknown tag: '%s'\n", argv[i]);
      print_usage(stderr, argv);
      exit(1);
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int get_num_class()
{
  char buffer[MAX_BUF];
  int num_class;

  num_class = 0;
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (strncmp(buffer, "# Classe", 8) == 0) {
      num_class++;
    }
    if ((strncmp(buffer, "# Loading data :", 16) == 0 ||
        strncmp(buffer, "################", 16) == 0) && num_class != 0) {
      return num_class;
    }
  }

  if (num_class < 1) {
    fprintf(stderr, "Error: unable to read the number of classes !\n");
    exit(1);
  }
  return num_class;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void get_mat(Data *data)
{
  char buffer[MAX_BUF], *sp;
  int i, j;

  i = 0;
  while (fgets(buffer, MAX_BUF, stdin) != NULL &&
      strncmp(buffer, "# Loading data :", 16) != 0 &&
      strncmp(buffer, "################", 16) != 0) {
      if (strncmp(buffer, "  -> False positives", 20) == 0) {
      sp = buffer;
      for (j = 0; j < data->num_class; j++) {
        sp = strchr(sp, '[');
        sp++;
        data->mat[i][j] = atoi(sp);
      }
      i++;
    }
  }

/* calculate totals */

  memset(data->total_class, 0, data->num_class * sizeof(int));
  memset(data->total_obs, 0, data->num_class * sizeof(int));
  data->total = 0;
  for (i = 0; i < data->num_class; i++) {
    for (j = 0; j < data->num_class; j++) {
      data->total_class[j] += data->mat[i][j];
      data->total_obs[i] += data->mat[i][j];
      data->total += data->mat[i][j];
    }
  }
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define WIDTH 12

void print_latex_graph(Data *data)
{
  int i, j;
  float mid_x, mid_y;

  printf("\\begin{center}\n");
  printf("\\begin{pspicture}(%d,%d)\n", WIDTH, data->num_class);
  printf("\\rput[l](1,%d){\\color{grey4} 10--20~\\%%}\n",
      data->num_class + 1);
  printf("\\rput[l](3,%d){\\color{lightblue} 20--30~\\%%}\n",
      data->num_class + 1);
  printf("\\rput[l](5,%d){\\color{orange} 30--40~\\%%}\n",
      data->num_class + 1);
  printf("\\rput[l](7,%d){\\color{red} \\textbf{$>$ 40~\\%%}}\n",
      data->num_class + 1);
  printf("\\rput[r](0.5,%d){From}\n", data->num_class);
  printf("\\rput[l](%.1f,%d){To}\n", WIDTH - 0.5, data->num_class);

  mid_x = (float) WIDTH / 2.0;
  for (i = 0; i < data->num_class; i++) {
    printf("\\psdot(1,%d)\\psdot(%d,%d)\n", i, WIDTH - 1, i);
    printf("\\rput[r](0.5,%d){%c~~(%.1f~\\%%)}"
        "\\rput[l](%.1f,%d){%c~~(%.1f~\\%%)}\n", i, AA1[i],
        100.0 * (float) data->total_class[i] / (float) data->total,
        WIDTH - 0.5, i, AA1[i],
        100.0 * (float) data->total_obs[i] / (float) data->total);

/* IMPORTANT: loops are repeated for placing more important arrows on top */

    for (j = 0; j < data->num_class; j++) {
      mid_y = (float) j + (float) (i - j) / 2.0;
      if (i != j && data->mat[i][j] != 0) {
        /* FROM j */
        if ((float) data->mat[i][j] / data->total_class[j] >= 0.1 &&
            (float) data->mat[i][j] / data->total_class[j] < 0.2) {
          printf("\\psline[linecolor=grey4](%d,%d)(%.1f,%.1f)\n",
              1, j, mid_x, mid_y);
        }
        /* TO i */
        if ((float) data->mat[i][j] / data->total_obs[i] >= 0.1 &&
            (float) data->mat[i][j] / data->total_obs[i] < 0.2) {
          printf("\\psline[linecolor=grey4]{->}(%.1f,%.1f)(%d,%d)\n",
              mid_x, mid_y, WIDTH - 1, i);
        }
      }
    }
  }
  for (i = 0; i < data->num_class; i++) {
    for (j = 0; j < data->num_class; j++) {
      mid_y = (float) j + (float) (i - j) / 2.0;
      if (i != j && data->mat[i][j] != 0) {
        /* FROM j */
        if ((float) data->mat[i][j] / data->total_class[j] >= 0.2 &&
            (float) data->mat[i][j] / data->total_class[j] < 0.3) {
          printf("\\psline[linecolor=lightblue](%d,%d)(%.1f,%.1f)\n",
              1, j, mid_x, mid_y);
        }
        /* TO i */
        if ((float) data->mat[i][j] / data->total_obs[i] >= 0.2 &&
            (float) data->mat[i][j] / data->total_obs[i] < 0.3) {
          printf("\\psline[linecolor=lightblue]{->}(%.1f,%.1f)(%d,%d)\n",
              mid_x, mid_y, WIDTH - 1, i);
        }
      }
    }
  }
  for (i = 0; i < data->num_class; i++) {
    for (j = 0; j < data->num_class; j++) {
      mid_y = (float) j + (float) (i - j) / 2.0;
      if (i != j && data->mat[i][j] != 0) {
        /* FROM j */
        if ((float) data->mat[i][j] / data->total_class[j] >= 0.3 &&
            (float) data->mat[i][j] / data->total_class[j] < 0.4) {
          printf("\\psline[linecolor=orange](%d,%d)(%.1f,%.1f)\n",
              1, j, mid_x, mid_y);
        }
        /* TO i */
        if ((float) data->mat[i][j] / data->total_obs[i] >= 0.3 &&
            (float) data->mat[i][j] / data->total_obs[i] < 0.4) {
          printf("\\psline[linecolor=orange]{->}(%.1f,%.1f)(%d,%d)\n",
              mid_x, mid_y, WIDTH - 1, i);
        }
      }
    }
  }
  for (i = 0; i < data->num_class; i++) {
    for (j = 0; j < data->num_class; j++) {
      mid_y = (float) j + (float) (i - j) / 2.0;
      if (i != j && data->mat[i][j] != 0) {
        /* FROM j */
        if ((float) data->mat[i][j] / data->total_class[j] >= 0.4) {
          printf("\\psline[linecolor=red, linewidth=.1](%d,%d)(%.1f,%.1f)\n",
              1, j, mid_x, mid_y);
        }
        /* TO i */
        if ((float) data->mat[i][j] / data->total_obs[i] >= 0.4) {
          printf("\\psline[linecolor=red, linewidth=.1]{->}(%.1f,%.1f)(%d,%d)\n",
              mid_x, mid_y, WIDTH - 1, i);
        }
      }
    }
  }
  printf("\\end{pspicture}\n");
  printf("\\end{center}\n\n");
}

/******************************************************************************
 *                                                                            *
 ******************************************************************************/
/**
 * Interchange v[i] and v[j]
 */
void swap(int *v, const int i, const int j)
{
  int temp;

  temp = v[i];
  v[i] = v[j];
  v[j] = temp;
}

/******************************************************************************
 *                                                                            *
 ******************************************************************************/
/**
 * Sort v[left]...v[right] in increasing order.
 * Note that array v itself is left unmodified, only the pointers to
 * the elements of v: index, are really sorted.
 *
 * WARNING: index should be initialized with increasing values from
 *          'left' to 'right' before.
 */
void fqsort(const float *val, int *index, const int left, const int right)
{
  int i, last;

  if (left >= right) {
    return;
  }
  swap(index, left, (left + right) / 2);
  last = left;
  for (i = left + 1; i <= right; i++) {
    if (val[index[i]] > val[index[left]]) {
      swap(index, ++last, i);
    }
  }
  swap(index, left, last);
  fqsort(val, index, left, last - 1);
  fqsort(val, index, last + 1, right);
  return;
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/

#define ARRAY_WIDTH 12

void print_latex_array(Data *data, int flag)
{
  int i, j, *index;
  float beg, fact, *val, tmp;

  ALLOC(val, data->num_class, float);
  ALLOC(index, data->num_class, int);
  fact = 100.0 / ARRAY_WIDTH;

  printf("\n\\begin{center}\n");
  printf("\\begin{pspicture}(%d,%d)\n", ARRAY_WIDTH, data->num_class);
  for (i = 0; i < data->num_class; i++) {
    if (flag == 0) {
      tmp = (float) data->total_class[i];
    } else {
      tmp = (float) data->total_obs[i];
    }
    printf("\\rput[l](%.2f,%.2f){%c~~(%.2f~\\%%)}\n", 0.0, (float) i + 0.5,
        AA1[i], 100.0 * tmp / (float) data->total);

    for (j = 0; j < data->num_class; j++) {
      if (flag == 0) {
        val[j] = fact * (float) data->mat[j][i] / data->total_class[i];
      } else {
        val[j] = fact * (float) data->mat[i][j] / data->total_obs[i];
      }
      index[j] = j;
    }
    fqsort(val, index, 0, data->num_class - 1);

    beg = 2.5;
    for (j = 0; j < data->num_class; j++) {
      printf("\\psframe(%.2f,%d)(%.2f,%d)\n", beg, i, beg + val[index[j]],
          i + 1);
      if (val[index[j]] >= 0.5) {
        printf("\\rput(%.2f,%.2f){%c}\n", beg + val[index[j]] / 2.0,
            (float) i + 0.5, AA1[index[j]]);
      }
      beg += val[index[j]];
    }
  }
  printf("\\end{pspicture}\n");
  printf("\\end{center}\n\n");

  FREE(val);
  FREE(index);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
void dump_matrix(Data *data, Parameters *params)
{
  int i, j;
  FILE *fp;

  FOPEN(fp, params->dump_file_name, "a");
  for (i = 0; i < data->num_class; i++) {
    for (j = 0; j < data->num_class; j++) {
      fprintf(fp, "%7.1f", -100.0 * ((float) data->mat[i][j] /
          data->total_class[j]));
    }
    fprintf(fp, "\n");
  }
  FCLOSE(fp);
}

/******************************************************************************/
/*                                                                            */
/******************************************************************************/
int main(int argc, char **argv)
{
  char buffer[MAX_BUF];
  Parameters params;
  Data data;

  init_parameters(&params);
  process_command_line(argc, argv, &params);

/* get number of classes */

  data.num_class = get_num_class();

  ALLOC(data.total_class, data.num_class, int);
  ALLOC(data.total_obs, data.num_class, int);
  ALLOC2(data.mat, data.num_class, data.num_class, int);

  rewind(stdin);
  while (fgets(buffer, MAX_BUF, stdin) != NULL) {
    if (strncmp(buffer, "# Classe 0", 10) == 0) {

/* get counts */

      get_mat(&data);

/* print graph */

      print_latex_graph(&data);
      print_latex_array(&data, 0);
      print_latex_array(&data, 1);

      if (params.dump_matrix == 1) {
        dump_matrix(&data, &params);
      }
    }
  }

  FREE(data.total_class);
  FREE(data.total_obs);
  FREE2(data.mat);

  return 0;
}
